import numpy as np
import torch

from neuralogic.dataset import Data


def test_data_from_python_lists() -> None:
    """Tests data creation from python lists and scalars"""
    features = [[0, 1], [1, 0], [1, 0], [0, 1]]
    edges = [[0, 2], [1, 3]]

    y = 1

    data = Data(x=features, edge_index=edges, y=y)
    query, example = data.to_logic_form()

    expected_examples = [
        "<1> edge(0, 1).",
        "<1> edge(2, 3).",
        "<[0, 1]> node_feature(0).",
        "<[1, 0]> node_feature(1).",
        "<[1, 0]> node_feature(2).",
        "<[0, 1]> node_feature(3).",
    ]

    assert str(query) == "1.0 predict."
    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)

    y = [2]

    data = Data(x=features, edge_index=edges, y=y)
    query, _ = data.to_logic_form()
    assert str(query) == "2.0 predict."

    query, _ = data.to_logic_form(one_hot_encode_labels=True, max_classes=3)
    assert str(query) == "[0.0, 0.0, 1.0] predict."

    y = [0, 1]

    data = Data(x=features, edge_index=edges, y=y)
    query, example = data.to_logic_form(
        feature_name="feat", edge_name="bond", output_name="out", one_hot_decode_features=True
    )

    expected_examples = [
        "<1> bond(0, 1).",
        "<1> bond(2, 3).",
        "<1> feat_1(0).",
        "<1> feat_0(1).",
        "<1> feat_0(2).",
        "<1> feat_1(3).",
    ]

    assert str(query) == "[0, 1] out."
    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)

    data = Data(x=features, edge_index=edges, y=[0, 1, 2, 3], y_mask=[1, 2])
    query, _ = data.to_logic_form()

    assert len(query) == 2
    assert str(query[0]) == "1.0 predict(1)."
    assert str(query[1]) == "2.0 predict(2)."


def test_data_from_numpy() -> None:
    """Tests data creation from numpy"""
    features = np.array([[0, 1], [1, 0], [1, 0], [0, 1]])
    edges = np.array([[0, 2], [1, 3]])

    y = np.array(1)

    data = Data(x=features, edge_index=edges, y=y)
    query, example = data.to_logic_form()

    expected_examples = [
        "<1> edge(0, 1).",
        "<1> edge(2, 3).",
        "<[0, 1]> node_feature(0).",
        "<[1, 0]> node_feature(1).",
        "<[1, 0]> node_feature(2).",
        "<[0, 1]> node_feature(3).",
    ]

    assert str(query) == "1.0 predict."
    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)

    y = np.array([2])

    data = Data(x=features, edge_index=edges, y=y)
    query, _ = data.to_logic_form()
    assert str(query) == "2.0 predict."

    query, _ = data.to_logic_form(one_hot_encode_labels=True, max_classes=3)
    assert str(query) == "[0.0, 0.0, 1.0] predict."

    y = np.array([0, 1])

    data = Data(x=features, edge_index=edges, y=y)
    query, example = data.to_logic_form(
        feature_name="feat", edge_name="bond", output_name="out", one_hot_decode_features=True
    )

    expected_examples = [
        "<1> bond(0, 1).",
        "<1> bond(2, 3).",
        "<1> feat_1(0).",
        "<1> feat_0(1).",
        "<1> feat_0(2).",
        "<1> feat_1(3).",
    ]

    assert str(query) == "[0, 1] out."
    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)

    data = Data(x=features, edge_index=edges, y=np.array([0, 1, 2, 3]), y_mask=np.array([1, 2]))
    query, _ = data.to_logic_form()

    assert len(query) == 2
    assert str(query[0]) == "1.0 predict(1)."
    assert str(query[1]) == "2.0 predict(2)."


def test_data_from_torch() -> None:
    """Tests data creation from torch"""
    features = torch.tensor([[0, 1], [1, 0], [1, 0], [0, 1]])
    edges = torch.tensor([[0, 2], [1, 3]])

    y = torch.tensor(1)

    data = Data(x=features, edge_index=edges, y=y)
    query, example = data.to_logic_form()

    expected_examples = [
        "<1> edge(0, 1).",
        "<1> edge(2, 3).",
        "<[0, 1]> node_feature(0).",
        "<[1, 0]> node_feature(1).",
        "<[1, 0]> node_feature(2).",
        "<[0, 1]> node_feature(3).",
    ]

    assert str(query) == "1.0 predict."
    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)

    y = torch.tensor([2])

    data = Data(x=features, edge_index=edges, y=y)
    query, _ = data.to_logic_form()
    assert str(query) == "2.0 predict."

    query, _ = data.to_logic_form(one_hot_encode_labels=True, max_classes=3)
    assert str(query) == "[0.0, 0.0, 1.0] predict."

    y = torch.tensor([0, 1])

    data = Data(x=features, edge_index=edges, y=y)
    query, example = data.to_logic_form(
        feature_name="feat", edge_name="bond", output_name="out", one_hot_decode_features=True
    )

    expected_examples = [
        "<1> bond(0, 1).",
        "<1> bond(2, 3).",
        "<1> feat_1(0).",
        "<1> feat_0(1).",
        "<1> feat_0(2).",
        "<1> feat_1(3).",
    ]

    assert str(query) == "[0, 1] out."
    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)

    data = Data(x=features, edge_index=edges, y=torch.tensor([0, 1, 2, 3]), y_mask=torch.tensor([1, 2]))
    query, _ = data.to_logic_form()

    assert len(query) == 2
    assert str(query[0]) == "1.0 predict(1)."
    assert str(query[1]) == "2.0 predict(2)."


def test_data_edge_features_from_torch() -> None:
    """Tests data creation with edge features from torch"""

    features = torch.tensor([[0, 1], [1, 0], [1, 0], [0, 1]])
    edges = torch.tensor([[0, 2], [1, 3]])
    edge_features = torch.tensor([[0, 1, 2], [3, 4, 5]])

    data = Data(x=features, edge_index=edges, edge_attr=edge_features, y=0)
    query, example = data.to_logic_form()

    expected_examples = [
        "<[0, 1, 2]> edge(0, 1).",
        "<[3, 4, 5]> edge(2, 3).",
        "<[0, 1]> node_feature(0).",
        "<[1, 0]> node_feature(1).",
        "<[1, 0]> node_feature(2).",
        "<[0, 1]> node_feature(3).",
    ]

    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)


def test_data_edge_features_types() -> None:
    """Tests data creation with edge features from torch"""

    features = torch.tensor([])
    edges = torch.tensor([[0, 2], [1, 3]])
    edge_features = torch.tensor([[0, 1, 0], [1, 0, 0]])

    data = Data(x=features, edge_index=edges, edge_attr=edge_features, y=0)
    query, example = data.to_logic_form(one_hot_decode_edge_features=True)

    expected_examples = [
        "<1> edge_1(0, 1).",
        "<1> edge_0(2, 3).",
    ]

    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)


def test_data_edge_features_from_numpy() -> None:
    """Tests data creation with edge features from numpy"""

    features = np.array([[0, 1], [1, 0], [1, 0], [0, 1]])
    edges = np.array([[0, 2], [1, 3]])
    edge_features = np.array([[0, 1, 2], [3, 4, 5]])

    data = Data(x=features, edge_index=edges, edge_attr=edge_features, y=0)
    query, example = data.to_logic_form()

    expected_examples = [
        "<[0, 1, 2]> edge(0, 1).",
        "<[3, 4, 5]> edge(2, 3).",
        "<[0, 1]> node_feature(0).",
        "<[1, 0]> node_feature(1).",
        "<[1, 0]> node_feature(2).",
        "<[0, 1]> node_feature(3).",
    ]

    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)


def test_data_edge_features_from_python_lists() -> None:
    """Tests data creation with edge features from python lists"""

    features = [[0, 1], [1, 0], [1, 0], [0, 1]]
    edges = [[0, 2], [1, 3]]
    edge_features = [[0, 1, 2], [3, 4, 5]]

    data = Data(x=features, edge_index=edges, edge_attr=edge_features, y=0)
    query, example = data.to_logic_form()

    expected_examples = [
        "<[0, 1, 2]> edge(0, 1).",
        "<[3, 4, 5]> edge(2, 3).",
        "<[0, 1]> node_feature(0).",
        "<[1, 0]> node_feature(1).",
        "<[1, 0]> node_feature(2).",
        "<[0, 1]> node_feature(3).",
    ]

    for a, b in zip(expected_examples, example):
        assert str(a) == str(b)



from typing import Optional, List


class Variable:
    __slots__ = "name", "type"

    def __init__(self, name: str, type: Optional[str] = None):
        self.name = name
        self.type = type

    def __str__(self) -> str:
        if self.type is not None:
            return f"{self.type}:{self.name}"
        return f"{self.name}"

    def __getitem__(self, item) -> List["Variable"]:
        if not isinstance(item, slice):
            raise ValueError("Variable range can be only defined by a slice")

        if item.step is not None:
            return [Variable(f"{self.name}{i}", self.type) for i in range(item.start, item.stop, item.step)]

        return [Variable(f"{self.name}{i}", self.type) for i in range(item.start, item.stop)]

    def __call__(self, item: str) -> "Variable":
        if isinstance(item, str):
            return Variable(self.name, item)
        raise ValueError("Type can be only of type str")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (Variable, str)):
            raise NotImplementedError
        return str(other) == str(self)

    def __hash__(self):
        return str(self).__hash__()

    def __lt__(self, other):
        return str(self).__lt__(str(other))


class Constant:
    __slots__ = "name", "type"

    def __init__(self, name: str, type: Optional[str] = None):
        self.name = name
        self.type = type

    def __str__(self) -> str:
        if self.type is not None:
            return f"{self.type}:{self.name}"
        return f"{self.name}"

    def __call__(self, item: str) -> "Constant":
        if isinstance(item, str):
            return Constant(self.name, item)
        raise ValueError("Type can be only of type str")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (Constant, str)):
            raise NotImplementedError
        return str(other) == str(self)

    def __hash__(self):
        return str(self).__hash__()

    def __lt__(self, other):
        return str(self).__lt__(str(other))



from typing import Iterable

import jpype

from neuralogic.dataset import Dataset
from neuralogic.setup import is_initialized, initialize
from neuralogic.core.builder import Builder, DatasetBuilder
from neuralogic.core.constructs.relation import BaseRelation, WeightedRelation
from neuralogic.core.constructs.rule import Rule
from neuralogic.core.constructs.predicate import PredicateMetadata
from neuralogic.core.constructs.java_objects import JavaFactory
from neuralogic.core.constructs.factories import R
from neuralogic.core.neural_module import NeuralModule
from neuralogic.core.settings import SettingsProxy, Settings
from neuralogic.nn.module.module import Module


TemplateEntries = BaseRelation | WeightedRelation | Rule | PredicateMetadata


class Template(NeuralModule):
    def __init__(self, *, template_file: str | None = None):
        super().__init__()
        self._template: list[TemplateEntries] = []
        self._template_file = template_file

    def add_rule(self, rule) -> None:
        """Adds one rule to the template

        :param rule:
        :return:
        """
        self.add_rules([rule])

    def add_rules(self, rules: list[TemplateEntries]) -> None:
        """Adds multiple rules to the template

        :param rules:
        :return:
        """
        if self._neural_model is not None:
            raise ValueError("Cannot modify built template")
        self._parsed_template = None
        self._template.extend(rules)

    def add_module(self, module: Module):
        """Expands the module into rules and adds them into the template

        :param module:
        :return:
        """
        self.add_rules(module())

    def build(self, settings: Settings | None = None, torch: bool = False) -> "Template":
        java_factory = JavaFactory()
        settings_proxy = settings.create_proxy() if settings is not None else Settings().create_disconnected_proxy()

        parsed_template = self._get_parsed_template(settings_proxy, java_factory)
        neural_model = Builder(settings_proxy).build_model(parsed_template, settings_proxy)

        self._initialize_neural_module(
            DatasetBuilder(parsed_template, java_factory),
            settings_proxy,
            neural_model,
            torch,
        )

        return self

    def remove_duplicates(self):
        """Remove duplicates from the template"""
        if self._neural_model is not None:
            raise ValueError("Cannot modify built template")

        self._parsed_template = None

        entries = set()
        deduplicated_template: list[TemplateEntries] = []

        for entry in self._template:
            entry_str = str(entry)

            if entry_str in entries:
                continue
            entries.add(entry_str)
            deduplicated_template.append(entry)
        self._template = deduplicated_template

    def _get_parsed_template(self, settings: SettingsProxy, java_factory: JavaFactory):
        if not is_initialized():
            initialize()

        if self._parsed_template is not None:
            return self._parsed_template

        if self._template_file is not None:
            self._parsed_template = Builder(settings).build_template_from_file(settings, self._template_file)
            return self._parsed_template

        predicate_metadata = []
        weighted_rules = []
        valued_facts = []

        for rule in self._template:
            if isinstance(rule, PredicateMetadata):
                predicate_metadata.append(java_factory.get_predicate_metadata_pair(rule))
            elif isinstance(rule, Rule):
                weighted_rules.append(java_factory.get_rule(rule))
            elif isinstance(rule, (WeightedRelation, BaseRelation)):
                valued_facts.append(java_factory.get_valued_fact(rule, java_factory.get_variable_factory()))

        parsed_template = jpype.JClass("cz.cvut.fel.ida.logic.constructs.template.types.ParsedTemplate")
        template = parsed_template(jpype.java.util.ArrayList(weighted_rules), jpype.java.util.ArrayList(valued_facts))

        template.weightsMetadata = (jpype.java.util.List) @ jpype.java.util.ArrayList([])
        template.predicatesMetadata = jpype.java.util.ArrayList(predicate_metadata)

        metadata_processor = jpype.JClass("cz.cvut.fel.ida.logic.constructs.template.transforming.MetadataProcessor")
        metadata_processor = metadata_processor(settings.settings)

        metadata_processor.processMetadata(template)
        self._parsed_template = template

        return self._parsed_template

    def derivable_queries(self, example: list[BaseRelation | Rule] | None = None):
        settings = Settings(iso_value_compression=False, chain_pruning=False).create_disconnected_proxy()
        java_factory = JavaFactory()

        parsed_template = self._get_parsed_template(settings, java_factory)
        dataset_builder = DatasetBuilder(parsed_template, java_factory)

        try:
            grounded_dataset = dataset_builder.ground_dataset(Dataset().add(None, example), settings)
        except Exception:
            return {}

        results = [
            R.get(name)(sub)
            for sample in grounded_dataset
            for name, substitution in sample.atoms.items()
            for sub in substitution.keys()
        ]

        if not results:
            return {}
        return results

    def query(self, query: BaseRelation, examples: list[BaseRelation | Rule] | None = None):
        settings = Settings(iso_value_compression=False, chain_pruning=False).create_disconnected_proxy()
        java_factory = JavaFactory()

        parsed_template = self._get_parsed_template(settings, java_factory)
        dataset_builder = DatasetBuilder(parsed_template, java_factory)

        try:
            grounded_dataset = dataset_builder.ground_dataset(Dataset().add(query, examples), settings)
        except Exception:
            return {}

        results = [node.substitutions for sample in grounded_dataset for node in sample.get_atoms(query)]

        if not results:
            return {}
        return results

    def q(self, query: BaseRelation, examples: list[BaseRelation | Rule] | None = None):
        return self.query(query, examples)

    def __str__(self) -> str:
        return "\n".join(str(r) for r in self._template)

    def __repr__(self) -> str:
        return self.__str__()

    def __iadd__(self, other) -> "Template":
        if self._neural_model is not None:
            raise ValueError("Cannot modify built template")
        self._parsed_template = None
        if isinstance(other, Iterable):
            self._template.extend(other)
        elif isinstance(other, Module):
            self._template.extend(other())
        else:
            self._template.append(other)
        return self

    def __getitem__(self, item) -> TemplateEntries:
        return self._template[item]

    def __delitem__(self, key):
        if self._neural_model is not None:
            raise ValueError("Cannot modify built template")
        self._parsed_template = None
        self._template.pop(key)

    def __setitem__(self, key, value):
        if self._neural_model is not None:
            raise ValueError("Cannot modify built template")
        if isinstance(value, (Iterable, Module)):
            raise NotImplementedError
        self._parsed_template = None
        self._template[key] = value

    def __len__(self) -> int:
        return len(self._template)

    def __iter__(self):
        return iter(self._template)

    def __copy__(self) -> "Template":
        temp = Template()

        temp._template_file = self._template_file
        temp._template = [rule for rule in self._template]

        return temp

    def clone(self) -> "Template":
        return self.__copy__()



from neuralogic.core import Template, Transformation, Aggregation
from neuralogic.nn.module import (
    RGCNConv,
    SAGEConv,
    GCNConv,
    GINConv,
    TAGConv,
    GATv2Conv,
    SGConv,
    APPNPConv,
    ResGatedGraphConv,
)


def test_rgcnconv():
    template = Template()

    template += RGCNConv(1, 2, "h1", "h0", "_edge", ["a", "b", "c"])
    template_str = str(template).split("\n")

    assert template_str[0] == "h1(I) :- {2, 1} h0(I). [aggregation=avg]"
    assert template_str[1] == "h1(I) :- {2, 1} h0(J), *edge(J, a, I). [aggregation=avg]"
    assert template_str[2] == "h1(I) :- {2, 1} h0(J), *edge(J, b, I). [aggregation=avg]"
    assert template_str[3] == "h1(I) :- {2, 1} h0(J), *edge(J, c, I). [aggregation=avg]"
    assert template_str[4] == "h1/1 [transformation=identity]"


def test_rgcnconv_relations_edge_replace():
    template = Template()

    template += RGCNConv(1, 2, "h1", "h0", None, ["a", "b", "c"], Transformation.SIGMOID)
    template_str = str(template).split("\n")

    assert template_str[0] == "h1(I) :- {2, 1} h0(I). [aggregation=avg]"
    assert template_str[1] == "h1(I) :- {2, 1} h0(J), a(J, I). [aggregation=avg]"
    assert template_str[2] == "h1(I) :- {2, 1} h0(J), b(J, I). [aggregation=avg]"
    assert template_str[3] == "h1(I) :- {2, 1} h0(J), c(J, I). [aggregation=avg]"
    assert template_str[4] == "h1/1 [transformation=sigmoid]"


def test_gcnconv():
    template = Template()

    template += GCNConv(1, 2, "h1", "h0", "edge")
    template_str = str(template).split("\n")

    assert template_str[0] == "<1.0> h1__edge(I, I)."
    assert template_str[1] == "h1__edge(I, J) :- edge(I, J)."
    assert template_str[2] == "h1__edge_count(I, J) :- h1__edge(J, X). [aggregation=count]"
    assert template_str[3] == "h1__edge_count(I, J) :- h1__edge(I, X). [aggregation=count]"
    assert template_str[4] == "h1__edge_count/2 [transformation=inverse, combination=product]"
    assert (
        template_str[5]
        == "{2, 1} h1(I) :- h0(J), h1__edge(J, I), sqrt(h1__edge_count(J, I)). [combination=product, aggregation=sum]"
    )


def test_sageconv():
    template = Template()

    template += SAGEConv(1, 2, "h1", "h0", "_edge")
    template_str = str(template).split("\n")

    assert template_str[0] == "{2, 1} h1(I) :- h0(J), *edge(J, I). [aggregation=avg]"
    assert template_str[1] == "{2, 1} h1(I) :- h0(I). [aggregation=avg]"


def test_tagconv():
    template = Template()

    template += TAGConv(1, 2, "h1", "h0", "_edge")
    template_str = str(template).split("\n")

    zero_hop = "{2, 1} h1(I0) :- h0(I0). [aggregation=sum]"
    sec_hop = "{2, 1} h1(I0) :- h0(I1), *edge(I1, I0). [aggregation=sum]"
    hop = "{2, 1} h1(I0) :- h0(I2), *edge(I1, I0), *edge(I2, I1). [aggregation=sum]"

    assert template_str[0] == zero_hop
    assert template_str[1] == sec_hop
    assert template_str[2] == hop

    template = Template()

    template += TAGConv(1, 2, "h1", "h0", "_edge", 1)
    template_str = str(template).split("\n")

    assert template_str[0] == zero_hop
    assert template_str[1] == sec_hop
    assert template_str[2] == "h1/1 [transformation=identity]"


def test_gatv2conv():
    template = Template()

    template += GATv2Conv(1, 2, "h1", "h0", "_edge")
    template_str = str(template).split("\n")

    attention = (
        "{2, 2} h1__attention(I, J) :- $h1__left={2, 1} h0(I), $h1__right={2, 1} h0(J). [transformation=leakyrelu]"
    )
    assert template_str[0] == attention
    assert template_str[1] == "h1__attention/2 [transformation=softmax]"

    h1_rule = (
        "h1(I) :- h1__attention(I, J), $h1__right={2, 1} h0(J), *edge(J, I). [combination=product, aggregation=sum]"
    )
    assert template_str[2] == h1_rule

    template = Template()

    template += GATv2Conv(1, 2, "h1", "h0", "_edge", share_weights=True)
    template_str = str(template).split("\n")

    attention = (
        "{2, 2} h1__attention(I, J) :- $h1__right={2, 1} h0(I), $h1__right={2, 1} h0(J). [transformation=leakyrelu]"
    )
    assert template_str[0] == attention
    assert template_str[1] == "h1__attention/2 [transformation=softmax]"

    h1_rule = (
        "h1(I) :- h1__attention(I, J), $h1__right={2, 1} h0(J), *edge(J, I). [combination=product, aggregation=sum]"
    )
    assert template_str[2] == h1_rule


def test_sgconv():
    template = Template()

    template += SGConv(1, 2, "h1", "h0", "_edge", k=2)
    template_str = str(template).split("\n")
    rule = "{2, 1} h1(I0) :- h0(I2), *edge(I1, I0), *edge(I2, I1). [aggregation=sum, duplicate_grounding=True]"

    assert template_str[0] == rule

    template = Template()

    template += SGConv(1, 2, "h1", "h0", "_edge")
    template_str = str(template).split("\n")
    rule = "{2, 1} h1(I0) :- h0(I1), *edge(I1, I0). [aggregation=sum, duplicate_grounding=True]"

    assert template_str[0] == rule


def test_appnp():
    template = Template()

    template += APPNPConv("h1", "h0", "_edge", 1, 0.1)
    template_str = str(template).split("\n")

    assert template_str[0] == "h1(I) :- <0.1> h0(I). [aggregation=sum]"
    assert template_str[1] == "h1(I) :- <0.9> h0(J), *edge(J, I). [aggregation=sum]"

    template = Template()

    template += APPNPConv("h1", "h0", "_edge", 3, 0.1)
    template_str = str(template).split("\n")

    assert template_str[0] == "h1__1(I) :- <0.1> h0(I). [aggregation=sum]"
    assert template_str[1] == "h1__1(I) :- <0.9> h0(J), *edge(J, I). [aggregation=sum]"

    assert template_str[2] == "h1__2(I) :- <0.1> h0(I). [aggregation=sum]"
    assert template_str[3] == "h1__2(I) :- <0.9> h1__1(J), *edge(J, I). [aggregation=sum]"

    assert template_str[4] == "h1(I) :- <0.1> h0(I). [aggregation=sum]"
    assert template_str[5] == "h1(I) :- <0.9> h1__2(J), *edge(J, I). [aggregation=sum]"


def test_res_gated():
    template = Template()

    template += ResGatedGraphConv(1, 2, "h1", "h0", "edge")
    template_str = str(template).split("\n")

    rule = "h1(I) :- h1__gate(I, J), {2, 1} h0(J), edge(J, I). [combination=elproduct, aggregation=sum]"

    assert template_str[0] == "h1__gate(I, J) :- {2, 1} h0(I), {2, 1} h0(J)."
    assert template_str[1] == "h1__gate/2 [transformation=sigmoid]"
    assert template_str[2] == "h1(I) :- {2, 1} h0(I)."
    assert template_str[3] == rule



import pytest

import torch

from neuralogic.core import Template, R, Settings, Transformation
from neuralogic.core.constructs.function.function import TransformationFunction
from neuralogic.nn.module import Linear
from neuralogic.dataset import Dataset, Sample


@pytest.mark.parametrize(
    "feature_size, output_size, num_of_inputs, activation",
    (
        (2, 3, 5, Transformation.IDENTITY),
        (3, 2, 2, Transformation.IDENTITY),
        (3, 3, 1, Transformation.IDENTITY),
        (3, 3, 1, Transformation.TANH),
    ),
)
def test_linear_module(feature_size: int, output_size: int, num_of_inputs: int, activation: TransformationFunction):
    """Test that PyNeuraLogic linear layer computes the same as pytorch linear layer"""
    torch.manual_seed(1)
    linear_input = torch.rand((num_of_inputs, feature_size))
    linear = torch.nn.Linear(feature_size, output_size, bias=False)
    linear_output = linear(linear_input)

    if activation == Transformation.TANH:
        linear_output = torch.tanh_(linear_output)

    template = Template()
    template += Linear(feature_size, output_size, "h", "f", activation=activation)
    model = template.build(Settings(iso_value_compression=False, chain_pruning=False))

    state = model.state_dict()
    state["weights"][0] = list(linear.parameters())[0].detach().numpy()
    model.load_state_dict(state)

    samples = [Sample(R.h(index), [R.f(index)[row.detach().numpy()]]) for index, row in enumerate(linear_input)]
    built_dataset = model.build_dataset(Dataset(samples))

    for sample, row in zip(built_dataset, linear_output):
        results = model.test(sample)

        assert len(results) == len(row)

        for a, b in zip(results, row):
            assert round(a, 3) == round(float(b), 3)



from neuralogic.utils.visualize import draw_model, draw_sample
from neuralogic.utils.data import XOR_Vectorized
from neuralogic.core import Settings


def test_draw_model():
    template, dataset = XOR_Vectorized()
    model = template.build(Settings())

    result = draw_model(model, show=False)

    assert isinstance(result, bytes)
    assert len(result) > 0


def test_draw_sample():
    template, dataset = XOR_Vectorized()
    model = template.build(Settings())

    built_dataset = model.build_dataset(dataset)
    result = draw_sample(built_dataset[0], show=False)

    assert isinstance(result, bytes)
    assert len(result) > 0


def test_draw_sample_from_raw_sample():
    template, dataset = XOR_Vectorized()
    model = template.build(Settings())

    built_dataset = model.build_dataset(dataset)
    result = built_dataset[0].draw(show=False)

    assert isinstance(result, bytes)
    assert len(result) > 0



from neuralogic.optim.lr_scheduler import LRDecay


class Optimizer:
    def __init__(self, lr: float, lr_decay: LRDecay | None = None):
        if lr_decay is not None:
            lr_decay._optimizer = self

        self._lr_decay = lr_decay
        self._lr = lr

        self._optimizer = None
        self._lr_object = None

    @property
    def lr(self) -> float:
        if self._lr_object is None:
            return self._lr
        return self._lr_object.value

    @lr.setter
    def lr(self, value: float):
        if self._lr_object is not None:
            self._lr_object.value = value
        self._lr = value

    def initialize(self):
        raise NotImplementedError

    def name(self) -> str:
        return self.__class__.__name__

    def get_lr_decay(self):
        if self._lr_decay is None:
            return None
        return self._lr_decay._initialize(self._lr_object)



import dataclasses
from collections import defaultdict
from typing import List, Dict, Tuple, Set

from neuralogic.core.constructs.relation import BaseRelation, WeightedRelation

from neuralogic.core import Aggregation, Settings, Metadata
from neuralogic.core.constructs.predicate import PredicateMetadata
from neuralogic.core.constructs.rule import Rule


@dataclasses.dataclass
class TableMapping:
    relation_name: str
    table_name: str
    term_columns: List[str]
    value_column: str | None = None


class Converter:
    def __init__(self, model, table_mappings: List[TableMapping], settings: Settings):
        self.table_mappings: Dict[str, TableMapping] = {
            f"{mapping.relation_name}/{len(mapping.term_columns)}": mapping for mapping in table_mappings
        }

        self.model = model
        self.settings = settings

        self._used_functions: Set[str] = set()

        self.sql_source = None
        self.std_functions = None

    def _process_template_entries(self) -> Tuple[defaultdict, Dict[str, Metadata]]:
        template = self.model.source_template
        weight_index = 0

        batched_relations: defaultdict[str, defaultdict[int, List]] = defaultdict(lambda: defaultdict(list))
        predicates_metadata = {}

        for entry in template:
            if isinstance(entry, Rule):
                weight_indices: List[int | None] = []
                if isinstance(entry.head, WeightedRelation) and entry.head.weight is not None:
                    weight_indices.append(weight_index)
                    weight_index += 1
                else:
                    weight_indices.append(None)
                for body_relation in entry.body:
                    if isinstance(body_relation, WeightedRelation) and body_relation.weight is not None:
                        weight_indices.append(weight_index)
                        weight_index += 1
                    else:
                        weight_indices.append(None)
                batched_relations[entry.head.predicate.name][entry.head.predicate.arity].append((entry, weight_indices))
            elif isinstance(entry, BaseRelation):
                if isinstance(entry, WeightedRelation) and entry.weight is not None:
                    weight_indices = [weight_index]
                    weight_index += 1
                else:
                    weight_indices = [None]
                batched_relations[entry.predicate.name][entry.predicate.arity].append((entry, weight_indices))
            elif isinstance(entry, PredicateMetadata):
                predicates_metadata[str(entry.predicate)] = entry.metadata
            else:
                raise NotImplementedError("Template can contain only relations or predicate metadata!")
        return batched_relations, predicates_metadata

    def _convert(self):
        weights = self.model.state_dict()["weights"]

        rule_default_activation = str(self.settings.rule_transformation).lower()
        relation_default_activation = str(self.settings.relation_transformation).lower()
        default_aggregation = str(Aggregation.AVG).lower()

        batched_relations, predicates_metadata = self._process_template_entries()

        sql_source_headers = []
        sql_source = []

        for name, arities in batched_relations.items():
            for arity, relations_by_arity in arities.items():
                if f"{name}/{arity}" in self.table_mappings:
                    raise Exception

                is_fact = False

                for index, (relation, weight_indices) in enumerate(relations_by_arity):
                    if isinstance(relation, Rule):
                        act, agg = rule_default_activation, default_aggregation

                        if relation.metadata is not None:
                            if relation.metadata.transformation is not None:
                                act = str(relation.metadata.transformation).lower()
                            if relation.metadata.aggregation is not None:
                                agg = str(relation.metadata.aggregation).lower()

                        self._used_functions.add(act)
                        self._used_functions.add(agg)

                        sql_func = self.get_rule_sql_function(relation, index, act, agg, weight_indices, weights)
                    else:
                        is_fact = True
                        sql_func = self.get_fact_sql_function(relation, index, weight_indices, weights)
                    sql_source.append(sql_func)

                activation = relation_default_activation
                aggregation = default_aggregation
                predicate_metadata = predicates_metadata.get(f"{name}/{arity}", None)

                if predicate_metadata is not None:
                    if predicate_metadata.transformation is not None:
                        activation = str(predicate_metadata.transformation).lower()
                    if predicate_metadata.aggregation is not None:
                        aggregation = str(predicate_metadata.aggregation).lower()

                self._used_functions.add(activation)
                self._used_functions.add(aggregation)

                sql_func = self.get_rule_aggregation_function(
                    name,
                    arity,
                    len(relations_by_arity),
                    activation,
                    aggregation,
                    is_fact,
                )

                sql_source.append(sql_func)

                sql_funcs = self.get_relation_interface_sql_function(name, arity)
                sql_source_headers.append(sql_funcs[0])
                sql_source.append(sql_funcs[1])

        self.std_functions = self.get_helpers(self._used_functions)

        sql_source_headers.extend(sql_source)
        self.sql_source = "\n".join(sql_source_headers)

    def get_relation_interface_sql_function(self, relation: str, arity: int) -> Tuple[str, str]:
        raise NotImplementedError

    def get_rule_sql_function(
        self, rule: Rule, index: int, activation: str, aggregation: str, weight_indices: List[int], weights
    ) -> str:
        raise NotImplementedError

    def get_fact_sql_function(self, relation: BaseRelation, index: int, weight_indices: List[int], weights) -> str:
        raise NotImplementedError

    def get_rule_aggregation_function(
        self, name: str, arity: int, number_of_rules: int, activation: str, aggregation: str, is_fact: bool = False
    ) -> str:
        raise NotImplementedError

    def get_helpers(self, functions: Set[str]) -> str:
        raise NotImplementedError

    def get_std_functions(self) -> str:
        if self.sql_source is None:
            self._convert()
        if self.sql_source is None:
            return ""

        return self.std_functions

    def to_sql(self) -> str:
        if self.sql_source is None:
            self._convert()
        if self.sql_source is None:
            return ""

        return self.sql_source

    @staticmethod
    def _is_var(term) -> bool:
        """Helper check if term is a variable or constant"""
        return str(term)[0].isupper()



from typing import Tuple

import jpype

from neuralogic.optim.lr_scheduler import LRDecay
from neuralogic.optim.optimizer import Optimizer


class Adam(Optimizer):
    def __init__(
        self,
        lr: float = 0.001,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-08,
        lr_decay: LRDecay | None = None,
    ):
        super().__init__(lr, lr_decay)
        self._betas = betas
        self._eps = eps

    @property
    def betas(self) -> Tuple[float, float]:
        return self._betas

    @property
    def eps(self) -> float:
        return self._eps

    def initialize(self):
        if self._optimizer:
            return self._optimizer

        adam_class = jpype.JClass("cz.cvut.fel.ida.neural.networks.computation.training.optimizers.Adam")
        self._lr_object = jpype.JClass("cz.cvut.fel.ida.algebra.values.ScalarValue")(self._lr)
        self._optimizer = adam_class(self._lr_object, self._betas[0], self._betas[1], self._eps)

        return self._optimizer

    def __str__(self) -> str:
        return f"Adam(lr={self.lr}, betas={self.betas}, eps={self.eps}, lr_decay={self._lr_decay})"



from typing import List
import itertools

import pytest

from neuralogic import manual_seed
from neuralogic.core import Settings, R, V, Template, Transformation
from neuralogic.dataset import Dataset, Sample
from neuralogic.optim import SGD


@pytest.mark.parametrize(
    "n,expected",
    [
        (2, [0, 1, 1, 0]),  # Number of inputs and expected output
        (3, [0, 1, 1, 0, 1, 0, 0, 1]),
        (4, [0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0]),
        (5, [0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1]),
    ],
)
def test_xor_generalization_accurate(n: int, expected: List[int]) -> None:
    manual_seed(0)

    dataset = Dataset()
    template = Template()

    template += (R.xor_at(0) <= R.val_at(0)) | [Transformation.TANH]
    template += (
        R.xor_at(V.Y)["a":1, 8] <= (R.val_at(V.Y)["b":8, 1], R.xor_at(V.X)["c":8, 1], R.special.next(V.X, V.Y))
    ) | [Transformation.TANH]

    dataset.add_samples(
        [
            Sample(R.xor_at(1)[0], [R.val_at(0)[0], R.val_at(1)[0]]),
            Sample(R.xor_at(1)[1], [R.val_at(0)[0], R.val_at(1)[1]]),
            Sample(R.xor_at(1)[1], [R.val_at(0)[1], R.val_at(1)[0]]),
            Sample(R.xor_at(1)[0], [R.val_at(0)[1], R.val_at(1)[1]]),
        ]
    )

    settings = Settings(epochs=5000)

    template.build(settings)
    template.train(dataset, epochs=5000)

    # build the dataset for n inputs
    products = itertools.product([0, 1], repeat=n)
    n_dataset = Dataset()

    for example in products:
        n_dataset.add_sample(Sample(R.xor_at(n - 1)[0], [R.val_at(i)[int(val)] for i, val in enumerate(example)]))

    for expected_value, predicted in zip(expected, template.test(n_dataset)):
        assert expected_value == predicted


@pytest.mark.parametrize(
    "n,expected",
    [
        (2, [0, 1, 1, 0]),  # Number of inputs and expected output
        (3, [0, 1, 1, 0, 1, 0, 0, 1]),
        (4, [0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0]),
    ],
)
def test_xor_generalization(n: int, expected: List[int]) -> None:
    """Tests xor generalization"""
    # fmt: off
    manual_seed(0)
    template = Template()

    # We have three weights in total named "a", "b" and "c"
    template.add_rules([

        # This rule does xor for the last pair
        (R.xor(V.X, V.Y)["a":1, 8] <= (
            R.x(V.X)["b":8, 1], R.x(V.Y)["c":8, 1], R.hidden.xy(V.X, V.Y), R.hidden.n(V.Y)
        )) | [Transformation.TANH],

        # This rule recursively evaluates xor for X and xor(Y, Z)
        (R.xor(V.X, V.Y)["a":1, 8] <= (
            R.x(V.X)["b":8, 1], R.xor(V.Y, V.Z)["c":8, 1], R.hidden.xy(V.X, V.Y), R.hidden.xy(V.Y, V.Z)
        )) | [Transformation.TANH],

        # Helper rule so that queries are just R.xor
        (R.xor <= R.xor(0, V.X)) | [Transformation.TANH],

        R.xor / 2 | [Transformation.TANH],
        R.xor / 0 | [Transformation.TANH],
    ])

    # The training dataset to train xor on two inputs x(0) and x(1), n(1) is means the max index of input is 1
    # x(0, 1) defines which input should be "xor-ed" together
    dataset = Dataset()
    dataset.add_samples([
        Sample(R.xor[0.0], [R.xy(0, 1), R.x(0)[0.0], R.x(1)[0.0], R.n(1)]),
        Sample(R.xor[1.0], [R.xy(0, 1), R.x(0)[1.0], R.x(1)[0.0], R.n(1)]),
        Sample(R.xor[1.0], [R.xy(0, 1), R.x(0)[0.0], R.x(1)[1.0], R.n(1)]),
        Sample(R.xor[0.0], [R.xy(0, 1), R.x(0)[1.0], R.x(1)[1.0], R.n(1)]),
    ])

    settings = Settings(optimizer=SGD())
    template.build(settings)

    # Train on the dataset with two var input
    template.train(dataset, epochs=300)

    # Get all products of lenght of n (all inputs of n vars)
    products = itertools.product([0.0, 1.0], repeat=n)

    n_dataset = Dataset()

    for example in products:
        # We make connections xy(0, 1), xy(1, 2), .. xy(n - 2, n - 1) and add them into the example
        fact_example = [R.xy(i, i + 1) for i in range(n - 1)]

        # Define values of individual variables (e.g., 0, 1, 0 -> R.x(0)[0], R.x(1)[1], R.x(2)[0]
        fact_example.extend([R.x(i)[v] for i, v in enumerate(example)])

        # Add info about the maximum index
        fact_example.append(R.n(n - 1))

        # Add example and query to the dataset, the query has some default value (1.0) as we do not care about the label
        n_dataset.add(R.xor, fact_example)

    # Check that we predicted correct values for n inputs for model trained on 2 inputs
    for expected_value, predicted in zip(expected, template.test(n_dataset)):
        assert expected_value == round(predicted)



import io
import os
import tempfile
from typing import Optional

import jpype

from neuralogic.setup import get_default_graphviz_path
from neuralogic.core.settings import Settings, SettingsProxy


def get_graphviz_path(path: Optional[str] = None) -> str | None:
    """
    Get the path to the Graphviz executable
    """
    if path is not None:
        return path
    return get_default_graphviz_path()


def get_drawing_settings(
    img_type: str = "png", value_detail: int = 0, graphviz_path: Optional[str] = None
) -> SettingsProxy:
    """Returns the default settings instance for drawing with a specified image type.

    :param img_type:
    :param value_detail:
    :param graphviz_path:
    :return:
    """
    settings = Settings().create_proxy()

    graphviz = get_graphviz_path(graphviz_path)
    if graphviz is not None:
        settings.settings.graphvizPath = graphviz

    settings.settings.drawing = False
    settings.settings.storeNotShow = True
    settings.settings.imgType = img_type.lower()
    settings.settings.outDir = tempfile.gettempdir()

    if value_detail not in [0, 1, 2]:
        raise ValueError(f"Invalid value_detail - {value_detail}. Expected 0, 1, or 2.")

    settings_class = settings.settings_class
    details = [
        settings_class.shortNumberFormat,
        settings_class.detailedNumberFormat,
        settings_class.superDetailedNumberFormat,
    ]

    settings.settings.defaultNumberFormat = details[value_detail]

    return settings


def get_template_drawer(settings: SettingsProxy):
    return jpype.JClass("cz.cvut.fel.ida.pipelines.debugging.drawing.TemplateDrawer")(settings.settings)


def get_sample_drawer(settings: SettingsProxy):
    return jpype.JClass("cz.cvut.fel.ida.pipelines.debugging.drawing.NeuralNetDrawer")(settings.settings)


def get_grounding_drawer(settings: SettingsProxy):
    return jpype.JClass("cz.cvut.fel.ida.pipelines.debugging.drawing.GroundingDrawer")(settings.settings)


# todo gusta: + groundingDrawer, pipelineDrawer...


def draw(drawer, obj, filename: Optional[str] = None, show=True, img_type="png", *args, **kwargs):
    if filename is not None:
        try:
            drawer.drawIntoFile(obj, os.path.abspath(filename))
        except jpype.java.lang.NullPointerException as e:
            raise Exception(
                "Drawing raised NullPointerException. Try to install GraphViz (https://graphviz.org/download/) on "
                "your Path or specify the path via the `graphviz_path` parameter"
            ) from e

        return None

    data = drawer.drawIntoBytes(obj)

    if data is None:
        raise Exception(
            "Drawing failed. Try to install GraphViz (https://graphviz.org/download/) on your Path or specify the "
            "path via the `graphviz_path` parameter"
        )

    data = bytes(data)

    if show:
        if is_jupyter():
            from IPython.display import Image, SVG

            if img_type.lower() == "svg":
                return SVG(data, *args, **kwargs)
            return Image(data, *args, **kwargs)
        else:
            import matplotlib.pyplot as plt
            import matplotlib.image as mpimg

            img = mpimg.imread(io.BytesIO(data), format=img_type)
            fig = plt.figure()

            if hasattr(fig.canvas, "set_window_title"):
                fig.canvas.set_window_title(kwargs.get("title", ""))

            ax = fig.add_axes((0, 0, 1, 1))
            ax.axis("off")
            ax.imshow(img)

            plt.show()
            return
    return data


def to_dot_source(drawer, obj) -> str:
    return str(drawer.getGraphSource(obj))


def draw_model(
    model,
    filename: Optional[str] = None,
    show=True,
    img_type="png",
    value_detail: int = 0,
    graphviz_path: Optional[str] = None,
    *args,
    **kwargs,
):
    """Draws model either as an image of type img_type either into:
        * a file - if filename is specified),
        * an IPython Image or Image popup - if show is True
        * or bytes otherwise

    :param model:
    :param filename:
    :param show:
    :param img_type:
    :param value_detail:
    :param graphviz_path:
    :param args:
    :param kwargs:
    :return:
    """
    if model._need_sync:
        model._sync_template()

    template = model._parsed_template
    template_drawer = get_template_drawer(get_drawing_settings(img_type, value_detail, graphviz_path))

    return draw(template_drawer, template, filename, show, img_type, *args, **kwargs)


def draw_grounding(
    grounding,
    filename: Optional[str] = None,
    show=True,
    img_type="png",
    value_detail: int = 0,
    graphviz_path: Optional[str] = None,
    *args,
    **kwargs,
):
    """Draws sample's grounding either as an image of type img_type either into:
        * a file - if filename is specified),
        * an IPython Image or Image popup - if show is True
        * or bytes otherwise

    :param sample:
    :param filename:
    :param show:
    :param img_type:
    :param value_detail:
    :param graphviz_path:
    :param args:
    :param kwargs:
    :return:
    """
    grounding_drawer = get_grounding_drawer(get_drawing_settings(img_type, value_detail, graphviz_path))

    return draw(grounding_drawer, grounding, filename, show, img_type, *args, **kwargs)


def draw_sample(
    sample,
    filename: Optional[str] = None,
    show=True,
    img_type="png",
    value_detail: int = 0,
    graphviz_path: Optional[str] = None,
    *args,
    **kwargs,
):
    """Draws sample either as an image of type img_type either into:
        * a file - if filename is specified),
        * an IPython Image or Image popup - if show is True
        * or bytes otherwise

    :param sample:
    :param filename:
    :param show:
    :param img_type:
    :param value_detail:
    :param graphviz_path:
    :param show:
    :param args:
    :param kwargs:
    :return:
    """
    draw_object = sample._java_sample

    sample_drawer = get_sample_drawer(get_drawing_settings(img_type, value_detail, graphviz_path))

    return draw(sample_drawer, draw_object, filename, show, img_type, *args, **kwargs)


def model_to_dot_source(model) -> str:
    """Renders the model into its dot source representation.

    :param model:
    :return:
    """
    if model._need_sync:
        model._sync_template()

    template = model._template
    template_drawer = get_template_drawer(get_drawing_settings())

    return to_dot_source(template_drawer, template)


def sample_to_dot_source(sample, value_detail: int = 0) -> str:
    """Renders the sample into its dot source representation.

    :param sample:
    :param value_detail:
    :return:
    """
    sample_drawer = get_sample_drawer(get_drawing_settings(value_detail=value_detail))

    return to_dot_source(sample_drawer, sample._java_sample)


def is_jupyter() -> bool:
    try:
        __IPYTHON__  # noqa: F821
        return True
    except Exception:
        return False



from typing import Optional

import jpype


class Function:
    __slots__ = "name", "operator", "can_flatten", "namespace"

    def __init__(self, name: str, *, namespace: str = "", operator: Optional[str] = None, can_flatten: bool = False):
        self.name: str = name.lower()
        self.operator: Optional[str] = operator
        self.can_flatten = can_flatten
        self.namespace = namespace

    def __str__(self):
        return self.name

    def wrap(self, content: str) -> str:
        return f"{self.name}({content})"

    def pretty_str(self) -> str:
        return str(self).capitalize()

    def __call__(self, *args, **kwargs):
        if len(args) == 0:
            return self
        raise NotImplementedError

    def is_parametrized(self) -> bool:
        return False

    def rule_head_dependant(self) -> bool:
        return False

    def process_head(self, head) -> "Function":
        raise NotImplementedError

    def get(self):
        name = "".join(s.capitalize() for s in self.name.split("_"))
        formatted_namespace = self.namespace.format(name=name)

        return jpype.JClass(f"cz.cvut.fel.ida.algebra.functions.{formatted_namespace}")()


class TransformationFunction(Function):
    def __init__(
        self,
        name: str,
        *,
        namespace: str = "transformation.elementwise.{name}",
        operator: Optional[str] = None,
        can_flatten: bool = False,
    ):
        super().__init__(name, namespace=namespace, operator=operator, can_flatten=can_flatten)

    def __call__(self, relation=None, **kwargs):
        from neuralogic.core.constructs import relation as rel
        from neuralogic.core.constructs.function.function_container import FContainer

        if relation is None:
            return self

        if isinstance(relation, rel.BaseRelation) and not isinstance(relation, rel.WeightedRelation):
            if relation.negated or relation.function is not None:
                return FContainer((relation,), self)
            return relation.attach_activation_function(self)
        return FContainer(relation, self)


class CombinationFunction(Function):
    def __init__(
        self,
        name: str,
        *,
        namespace: str = "combination.{name}",
        operator: Optional[str] = None,
        can_flatten: bool = False,
    ):
        super().__init__(name, namespace=namespace, operator=operator, can_flatten=can_flatten)

    def __call__(self, *relations):
        from neuralogic.core.constructs.function.function_container import FContainer

        if len(relations) == 0:
            return self
        return FContainer(relations, self)


class AggregationFunction(Function):
    def get(self):
        raise NotImplementedError



from typing import List

import jpype

from neuralogic.setup import is_initialized, initialize
from neuralogic.core.settings import SettingsProxy


class Sources:
    @staticmethod
    def from_settings(settings: SettingsProxy) -> "Sources":
        if not is_initialized():
            initialize()

        sources = jpype.JClass("cz.cvut.fel.ida.setup.Sources")(settings.settings)
        return Sources(sources)

    @staticmethod
    def from_args(args: List[str], settings: SettingsProxy) -> "Sources":
        if not is_initialized():
            initialize()

        runner = jpype.JClass("cz.cvut.fel.ida.neuralogic.cli.utils.Runner")
        sources = runner.getSources(args, settings.settings)
        settings._setup_random_generator()

        return Sources(sources)

    def __init__(self, sources):
        self.sources = sources

    def to_json(self) -> str:
        return self.sources.exportToJson()



from typing import Tuple
from types import EllipsisType

import jpype

from neuralogic.core.constructs.function.function import TransformationFunction


class Slice(TransformationFunction):
    __slots__ = ("rows", "cols")

    def __init__(
        self,
        name: str,
        *,
        rows: EllipsisType | Tuple[int, int] = ...,
        cols: EllipsisType | Tuple[int, int] = ...,
    ):
        super().__init__(name)

        self.cols = [int(x) for x in cols] if cols is not Ellipsis else Ellipsis
        self.rows = [int(x) for x in rows] if rows is not Ellipsis else Ellipsis

    def __call__(
        self,
        relation=None,
        *,
        rows: EllipsisType | Tuple[int, int] = ...,
        cols: EllipsisType | Tuple[int, int] = ...,
    ):
        slice = Slice(self.name, rows=rows, cols=cols)
        return TransformationFunction.__call__(slice, relation)

    def is_parametrized(self) -> bool:
        return True

    def get(self):
        cols = None if self.cols is Ellipsis else self.cols
        rows = None if self.rows is Ellipsis else self.rows

        return jpype.JClass("cz.cvut.fel.ida.algebra.functions.transformation.joint.Slice")(rows, cols)

    def wrap(self, content: str) -> str:
        rows = "..." if self.rows is Ellipsis or self.rows is None else self.rows
        cols = "..." if self.cols is Ellipsis or self.cols is None else self.cols

        return f"slice({content}, rows={rows}, cols={cols})"

    def __str__(self):
        rows = "..." if self.rows is Ellipsis or self.rows is None else self.rows
        cols = "..." if self.cols is Ellipsis or self.cols is None else self.cols

        return f"slice(rows={rows}, cols={cols})"



import os
from shutil import copyfileobj

from neuralogic.dataset.base import BaseDataset


class FileDataset(BaseDataset):
    r"""
    ``FileDataset`` represents samples stored in files in the `NeuraLogic <https://github.com/GustikS/NeuraLogic>`_
    (logic) format.

    Parameters
    ----------

    examples_file : str | None
        Path to the examples file. Default: ``None``
    queries_file : str | None
        Path to the queries file. Default: ``None``

    """

    def __init__(
        self,
        examples_file: str | None = None,
        queries_file: str | None = None,
    ):
        self.examples_file = examples_file
        self.queries_file = queries_file

        if self.examples_file is not None:
            self.examples_file = os.path.abspath(self.examples_file)
        if self.queries_file is not None:
            self.queries_file = os.path.abspath(self.queries_file)

    def dump(
        self,
        queries_fp,
        examples_fp,
        sep: str = "\n",
    ):
        if self.examples_file is not None:
            with open(self.examples_file, "r") as fp:
                copyfileobj(fp, examples_fp)

        if self.queries_file is not None:
            with open(self.queries_file, "r") as fp:
                copyfileobj(fp, queries_fp)



import jpype

from neuralogic.optim.lr_scheduler import LRDecay
from neuralogic.optim.optimizer import Optimizer


class SGD(Optimizer):
    def __init__(self, lr: float = 0.1, lr_decay: LRDecay | None = None):
        super().__init__(lr, lr_decay)

    def initialize(self):
        if self._optimizer:
            return self._optimizer

        sgd_class = jpype.JClass("cz.cvut.fel.ida.neural.networks.computation.training.optimizers.SGD")
        self._lr_object = jpype.JClass("cz.cvut.fel.ida.algebra.values.ScalarValue")(self._lr)
        self._optimizer = sgd_class(self._lr_object)

        return self._optimizer

    def __str__(self) -> str:
        return f"SGD(lr={self.lr}, lr_decay={self._lr_decay})"



import torch
from torch.nn import Sequential

import neuralogic
from neuralogic.core import Relation, Template, R, F
from neuralogic.nn.torch_function import NeuraLogic


def test_torch_function_with_parameters():
    torch.manual_seed(1)
    neuralogic.manual_seed(1)

    template = Template()
    template += Relation.xor[1, 8] <= Relation.xy

    def to_logic(tensor_data):
        return [Relation.xy[tensor_data]]

    torch_train_set = [
        (torch.tensor([0.0, 0.0]), torch.tensor(0.0)),
        (torch.tensor([0.0, 1.0]), torch.tensor(1.0)),
        (torch.tensor([1.0, 0.0]), torch.tensor(1.0)),
        (torch.tensor([1.0, 1.0]), torch.tensor(0.0)),
    ]

    sequential = Sequential(
        torch.nn.Linear(2, 8),
        torch.nn.Tanh(),
        NeuraLogic(
            template,
            [R.xy[8,]],
            R.xor,
            to_logic,
        ),
        torch.nn.Sigmoid(),
    )

    optimizer = torch.optim.SGD(sequential.parameters(), lr=0.1)
    loss = torch.nn.MSELoss()

    for _ in range(300):
        for input_data, label in torch_train_set:
            output = sequential(input_data)
            loss_value = loss(output, label)

            optimizer.zero_grad(set_to_none=True)
            loss_value.backward()
            optimizer.step()

    for input_data, label in torch_train_set:
        out = int(round(float(sequential(input_data))))
        float_label = int(label)

        assert out == float_label


def test_torch_function_without_parameters():
    torch.manual_seed(1)
    neuralogic.manual_seed(1)

    template = Template()
    template += Relation.xor <= Relation.xy

    def to_logic(tensor_data):
        return [Relation.xy[tensor_data]]

    torch_train_set = [
        (torch.tensor([0.0, 0.0]), torch.tensor(0.0)),
        (torch.tensor([0.0, 1.0]), torch.tensor(1.0)),
        (torch.tensor([1.0, 0.0]), torch.tensor(1.0)),
        (torch.tensor([1.0, 1.0]), torch.tensor(0.0)),
    ]

    sequential = Sequential(
        torch.nn.Linear(2, 8),
        torch.nn.Tanh(),
        NeuraLogic(
            template,
            [R.xy[8,]],
            R.xor,
            to_logic,
        ),
        torch.nn.Linear(8, 1),
        torch.nn.Sigmoid(),
    )

    optimizer = torch.optim.SGD(sequential.parameters(), lr=0.1)
    loss = torch.nn.MSELoss()

    for _ in range(400):
        for input_data, label in torch_train_set:
            output = sequential(input_data)
            loss_value = loss(output, label)

            optimizer.zero_grad(set_to_none=True)
            loss_value.backward()
            optimizer.step()

    for input_data, label in torch_train_set:
        out = int(round(float(sequential(input_data))))
        float_label = int(label)

        assert out == float_label



import jpype

from neuralogic.core.constructs.function.function import AggregationFunction, CombinationFunction


class ConcatCombination(CombinationFunction):
    __slots__ = ("axis",)

    def __init__(
        self,
        name: str,
        *,
        axis: int = -1,
    ):
        super().__init__(name)
        self.axis = axis

    def __call__(self, *relations, axis: int = -1):
        concat = ConcatCombination(self.name, axis=axis)
        return CombinationFunction.__call__(concat, *relations)

    def is_parametrized(self) -> bool:
        return self.axis != -1

    def get(self):
        return jpype.JClass("cz.cvut.fel.ida.algebra.functions.combination.Concatenation")(self.axis)

    def __str__(self):
        if self.axis == -1:
            return "concat"
        return f"concat(axis={self.axis})"


class ConcatAggregation(AggregationFunction):
    __slots__ = ("axis",)

    def __init__(
        self,
        name: str,
        *,
        axis: int = -1,
    ):
        super().__init__(name)
        self.axis = axis

    def __call__(self, *, axis: int = -1):
        concat = ConcatAggregation(self.name, axis=axis)
        return AggregationFunction.__call__(concat)

    def is_parametrized(self) -> bool:
        return self.axis != -1

    def get(self):
        return jpype.JClass("cz.cvut.fel.ida.algebra.functions.combination.Concatenation")(self.axis)

    def __str__(self):
        if self.axis == -1:
            return "concat"
        return f"concat(axis={self.axis})"



from neuralogic.optim.lr_scheduler.geometric import GeometricLR
from neuralogic.optim.lr_scheduler.arithmetic import ArithmeticLR
from neuralogic.optim.lr_scheduler.lr_decay import LRDecay


__all__ = ["ArithmeticLR", "GeometricLR", "LRDecay"]



import jpype
from tqdm.autonotebook import tqdm

from neuralogic.setup import is_initialized, initialize
from neuralogic.core.builder.components import NeuralSample
from neuralogic.core.settings import SettingsProxy
from neuralogic.core.sources import Sources


def stream_to_list(stream) -> list:
    return list(stream.collect(jpype.JClass("java.util.stream.Collectors").toList()))


class Builder:
    def __init__(self, settings: SettingsProxy):
        if not is_initialized():
            initialize()

        self.settings = settings
        self.example_builder = Builder.get_builders(settings)
        self.builder = Builder.get_builders(settings)

        self.neural_model = jpype.JClass("cz.cvut.fel.ida.neural.networks.computation.training.NeuralModel")
        self.collectors = jpype.JClass("java.util.stream.Collectors")

        @jpype.JImplements(jpype.JClass("java.util.function.IntConsumer"))
        class Callback:
            def __init__(self, progress_bar):
                self.state = 0
                self.progress_bar = progress_bar

            @jpype.JOverride
            def accept(self, count: int):
                self.state = max(count, self.state)
                if not self.progress_bar.disable:
                    self.progress_bar.update(1)

        self._callback = Callback

    def build_template_from_file(self, settings: SettingsProxy, filename: str):
        args = [
            "-t",
            filename,
            "-q",
            filename,
        ]

        sources = Sources.from_args(args, settings)
        template = self.builder.buildTemplate(sources.sources)

        return template

    def ground_from_sources(self, parsed_template, sources: Sources, progress: bool):
        if not progress:
            return self._ground(parsed_template, sources, None, None)
        with tqdm(total=None, desc="Grounding", unit=" samples", dynamic_ncols=True) as pbar:
            return self._ground(parsed_template, sources, None, self._callback(pbar))

    def ground_from_logic_samples(self, parsed_template, logic_samples, progress: bool):
        if not progress:
            return self._ground(parsed_template, None, logic_samples, None)
        with tqdm(total=len(logic_samples), desc="Grounding", unit=" samples", dynamic_ncols=True) as pbar:
            return self._ground(parsed_template, None, logic_samples, self._callback(pbar))

    def _ground(self, parsed_template, sources: Sources | None, logic_samples, callback):
        if sources is not None:
            ground_pipeline = self.example_builder.buildGroundings(parsed_template, sources.sources, callback)
        else:
            logic_samples = jpype.java.util.ArrayList(logic_samples).stream()
            ground_pipeline = self.example_builder.buildGroundings(parsed_template, logic_samples, callback)

        ground_pipeline.execute(None if sources is None else sources.sources)
        groundings = ground_pipeline.get()

        if callback is not None:
            return groundings.collect(self.collectors.toList())
        return groundings

    def neuralize(self, groundings, progress: bool, length: int | None) -> list[NeuralSample]:
        if not progress:
            return self._neuralize(groundings, None)
        with tqdm(total=length, desc="Building", unit=" samples", dynamic_ncols=True) as pbar:
            return self._neuralize(groundings, self._callback(pbar))

    def _neuralize(self, groundings, callback) -> list[NeuralSample]:
        neuralize_pipeline = self.example_builder.neuralize(groundings, callback)
        neuralize_pipeline.execute(None)
        logic_samples = neuralize_pipeline.get().collect(self.collectors.toList())

        return [NeuralSample(sample) for sample in logic_samples]

    def build_model(self, parsed_template, settings: SettingsProxy):
        neural_model = self.neural_model(parsed_template.getAllWeights(), settings.settings)

        return neural_model

    @staticmethod
    def get_builders(settings: SettingsProxy):
        builder = jpype.JClass("cz.cvut.fel.ida.pipelines.building.PythonBuilder")(settings.settings)

        return builder

    @staticmethod
    def _get_spinner_text(count: int) -> str:
        if count == 1:
            return f"Built {count} sample"
        return f"Built {count} samples"



from neuralogic.core.constructs.function.function import TransformationFunction, AggregationFunction
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class TAGConv(Module):
    r"""
    Topology Adaptive Graph Convolutional layer from
    `"Topology Adaptive Graph Convolutional Networks" <https://arxiv.org/abs/1710.10370>`_.
    Which can be expressed as:

    .. math::
        \mathbf{x}^{\prime}_i = act(\sum_{k=0}^K \mathbf{W}_k \cdot {agg}_{j \in \mathcal{N}^k(i)}(\mathbf{x}_j))

    Where *act* is an activation function, *agg* aggregation function, *Wk* are learnable parameters and
    :math:`\mathcal{N}^k(i)` denotes nodes that are *k* hops away from the node *i*. This equation is translated into
    the logic form as:

    This equation is translated into the logic form as:

    .. code:: logtalk

        (R.<output_name>(V.I0)[<W0>] <= R.<feature_name>(V.I0)) | [<aggregation>, Transformation.IDENTITY]
        (R.<output_name>(V.I0)[<W1>] <= (R.<feature_name>(V.I1), R.<edge_name>(V.I1, V.I0))) | [<aggregation>, Transformation.IDENTITY]
        (R.<output_name>(V.I0)[<W2>] <= (R.<feature_name>(V.I2), R.<edge_name>(V.I1, V.I0), R.<edge_name>(V.I2, V.I1)) | [<aggregation>, Transformation.IDENTITY]
        ...
        (R.<output_name>(V.I0)[<Wk>] <= (R.<feature_name>(V.I<k>), R.<edge_name>(V.I1, V.I0), ..., R.<edge_name>(V.I<k>, V.I<k-1>)) | [<aggregation>, Transformation.IDENTITY]
        R.<output_name> / 1 | [<activation>]

    Examples
    --------

    The whole computation of this module (parametrized as :code:`TAGConv(1, 2, "h1", "h0", "_edge")`) is as follows:

    .. code:: logtalk

        (R.h1(V.I0)[2, 2] <= R.h0(V.I0)) | [Aggregation.SUM, Transformation.IDENTITY]
        (R.h1(V.I0)[2, 1] <= (R.h0(V.I1), R._edge(V.I1, V.I0)) | [Aggregation.SUM, Transformation.IDENTITY]
        (R.h1(V.I0)[2, 1] <= (R.h0(V.I2), R._edge(V.I1, V.I0), R._edge(V.I2, V.I1)) | [Aggregation.SUM, Transformation.IDENTITY]
        R.h1 / 1 | [Transformation.IDENTITY]

    Module parametrized as :code:`TAGConv(1, 2, "h1", "h0", "_edge", 1)` translates into:

    .. code:: logtalk

        (R.h1(V.I0)[2, 1] <= R.h0(V.I0)) | [Aggregation.SUM, Transformation.IDENTITY]
        (R.h1(V.I0)[2, 1] <= (R.h0(V.I1), R._edge(V.I1, V.I0)) | [Aggregation.SUM, Transformation.IDENTITY]
        R.h1 / 1 | [Transformation.IDENTITY]

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : str
        Edge predicate name to use for neighborhood relations.
    k : int
        Number of hops.
        Default: ``2``
    activation : TransformationFunction
        Activation function of the output.
        Default: ``Transformation.IDENTITY``
    aggregation : AggregationFunction
        Aggregation function of nodes' neighbors.
        Default: ``Aggregation.SUM``

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: str,
        k: int = 2,
        activation: TransformationFunction = Transformation.IDENTITY,
        aggregation: AggregationFunction = Aggregation.SUM,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.k = k
        self.activation = activation
        self.aggregation = aggregation

    def __call__(self):
        metadata = Metadata(aggregation=self.aggregation)
        head = R.get(self.output_name)
        feature = R.get(self.feature_name)
        edge = R.get(self.edge_name)

        hop_rules = []

        for i in range(self.k + 1):
            hop_rules.append(
                (
                    head(V.I0)[self.out_channels, self.in_channels]
                    <= (
                        feature(f"I{i}"),
                        *(edge(f"I{b}", f"I{a}") for a, b in zip(range(i), range(1, i + 1))),
                    )
                )
                | metadata
            )

        return [
            *hop_rules,
            head / 1 | Metadata(transformation=self.activation),
        ]



from neuralogic.core.constructs.relation import BaseRelation, WeightedRelation
from neuralogic.core import R, Transformation, Aggregation, Metadata, Combination, V, C
from neuralogic.core.constructs.rule import Rule


def test_predicate_creation() -> None:
    """Test different ways of predicate creation"""
    relation = R.lorem_ipsum
    predicate = relation / 1

    assert predicate.arity == 1
    assert predicate.name == "lorem_ipsum"
    assert predicate.special is False
    assert predicate.hidden is False

    predicate = relation / 2
    assert predicate.arity == 2

    assert relation.predicate.arity == 0
    assert relation.predicate.name == "lorem_ipsum"
    assert relation.predicate.special is False
    assert relation.predicate.hidden is False

    relation = R.special.lorem_ipsum
    predicate = relation / 1

    assert predicate.special is True
    assert predicate.hidden is False
    assert relation.predicate.special is True
    assert relation.predicate.hidden is False

    relation = R.hidden.lorem_ipsum
    predicate = relation / 1

    assert predicate.special is False
    assert predicate.hidden is True
    assert relation.predicate.special is False
    assert relation.predicate.hidden is True

    relation = R.special.hidden.lorem_ipsum
    predicate = relation / 1

    assert predicate.special is True
    assert predicate.hidden is True
    assert relation.predicate.special is True
    assert relation.predicate.hidden is True

    relation = R.special._hidden_test
    predicate = relation / 1

    assert predicate.hidden is True
    assert predicate.name == "hidden_test"
    assert predicate.special is True

    relation = R._hidden_test
    predicate = relation / 1

    assert predicate.hidden is True
    assert predicate.special is False
    assert predicate.name == "hidden_test"

    relation = R.get("_hidden_test")
    predicate = relation / 1

    assert predicate.hidden is True
    assert predicate.special is False
    assert predicate.name == "hidden_test"

    predicate_metadata = R.shortest / 2 | [Transformation.SIGMOID]
    assert predicate_metadata.metadata is not None
    assert predicate_metadata.metadata.aggregation is None
    assert predicate_metadata.metadata.transformation == Transformation.SIGMOID

    predicate_metadata = R.shortest / 2 | [Transformation.SIGMOID, Combination.MAX]
    assert predicate_metadata.metadata is not None
    assert predicate_metadata.metadata.aggregation is None

    assert str(predicate_metadata.metadata.combination) == "max"
    assert str(predicate_metadata.metadata.transformation) == "sigmoid"

    predicate_metadata = R.shortest / 2 | [Combination.MIN]
    assert predicate_metadata.metadata is not None
    assert predicate_metadata.metadata.aggregation is None
    assert predicate_metadata.metadata.transformation is None

    assert str(predicate_metadata.metadata.combination) == "min"

    predicate_metadata = R.shortest / 2 | Metadata(transformation=Transformation.TANH, combination=Combination.MAX)
    assert predicate_metadata.metadata is not None
    assert predicate_metadata.metadata.aggregation == None

    assert str(predicate_metadata.metadata.combination) == "max"
    assert str(predicate_metadata.metadata.transformation) == "tanh"


def test_relation_creation() -> None:
    """Test relation creation related operations and properties"""
    relation = R.my_atom
    assert len(relation.terms) == 0
    assert isinstance(relation, BaseRelation)

    relation = relation("a", "b")
    assert len(relation.terms) == 2
    assert isinstance(relation, BaseRelation)

    relation = relation[1, 2]
    assert len(relation.terms) == 2
    assert isinstance(relation, WeightedRelation)
    assert relation.weight == (1, 2)

    relation = R.my_atom["abc":1, 2]
    assert len(relation.terms) == 0
    assert isinstance(relation, WeightedRelation)
    assert relation.weight == (1, 2)
    assert relation.weight_name == "abc"

    relation = relation.fixed()
    assert relation.is_fixed is True

    relation = R.my_atom
    neg_relation = -relation
    assert neg_relation.function is Transformation.REVERSE

    neg_relation = ~relation
    assert neg_relation.negated is True

    t_relation = relation.T
    assert t_relation.function is Transformation.TRANSP


def test_rule_metadata():
    rule = (R.a <= R.b) | [Transformation.SIGMOID, Aggregation.AVG]

    assert rule.metadata is not None
    assert rule.metadata.aggregation == Aggregation.AVG
    assert rule.metadata.transformation == Transformation.SIGMOID

    rule = (R.a <= R.b) | Metadata(transformation=Transformation.IDENTITY, aggregation=Aggregation.MAX)

    assert rule.metadata is not None
    assert rule.metadata.aggregation == Aggregation.MAX
    assert rule.metadata.transformation == Transformation.IDENTITY

    rule = R.a <= R.b
    assert rule.metadata is None

    rule = (R.a <= R.b) | [Combination.MAX, Transformation.SIGMOID, Aggregation.AVG]

    assert rule.metadata is not None
    assert rule.metadata.aggregation == Aggregation.AVG

    assert str(rule.metadata.combination) == "max"
    assert str(rule.metadata.transformation) == "sigmoid"


def test_rules():
    my_rule = R.a <= R.b

    assert len(my_rule.body) == 1
    assert my_rule.body[0].predicate.name == "b"


def test_rules_and():
    my_rule: Rule = R.a(V.X) <= R.b(V.Y) & R.c(V.Z) & R.d

    assert len(my_rule.body) == 3
    assert str(my_rule.body[0]) == "b(Y)."
    assert str(my_rule.body[1]) == "c(Z)."
    assert str(my_rule.body[2]) == "d."


def test_rules_and_with_metadata():
    my_rule: Rule = R.a(V.X) <= R.b(V.Y) & R.c(V.Z) & R.d | [Transformation.SIGMOID]

    assert str(my_rule) == "a(X) :- b(Y), c(Z), d. [transformation=sigmoid]"


def test_var_and_const():
    assert V.get("abc") == "Abc"
    assert V.get("abc", "type") == "type:Abc"

    assert C.get("Abc") == "abc"
    assert C.get("Abc", "type") == "type:abc"

    assert str(R.head(V.get("abc", "type")[1:3], V.X)) == "head(type:Abc1, type:Abc2, X)."

    # With step
    assert str(R.head(V.get("abc", "type")[1:6:2], V.X)) == "head(type:Abc1, type:Abc3, type:Abc5, X)."



class LRDecay:
    def __init__(self):
        self._optimizer = None
        self._decay = None

    def _initialize(self, learning_rate):
        raise NotImplementedError

    def _initialize_optimizer(self):
        if self._decay is not None:
            return
        if self._optimizer is None:
            raise ValueError("LRDecay cannot be initialized - no optimizer has been attached")
        self._optimizer.initialize()
        self._initialize(self._optimizer._lr_object)

    def decay(self, epoch: int):
        """
        Manually run the learning rate decay - this is useful when passing sample by sample into the training method
        instead of passing the whole batch of samples. In that case, the decay is not triggered automatically, as it
        is unknown what the current epoch is.

        Parameters
        ----------

        epoch : int
            The number of the current epoch.
        """
        self._initialize_optimizer()
        self._decay.decay(epoch)

    def restart(self):
        """
        Reset the learning rate, and the learning decay itself to its original state.
        """
        self._initialize_optimizer()
        self._decay.restart()



from typing import List, Union, Tuple, Sequence, Iterable

import numpy as np

from neuralogic.core.constructs.factories import Relation
from neuralogic.dataset.base import ConvertibleDataset
from neuralogic.dataset.logic import Dataset


class Data:
    r"""The ``Data`` instance stores information about one specific graph instance.

    Example
    -------

    For example, the directed graph :math:`G = (V, E)`, where :math:`E = \{(0, 1), (1, 2), (2, 0)\}`,
    node features :math:`X = \{[0], [1], [0]\}` and target nodes' labels
    :math:`Y = \{0, 1, 0\}` would be represented as:

    .. code-block:: python

        data = Data(
            x=[[0], [1], [0]],
            edge_index=[
                [0, 1, 2],
                [1, 2, 0],
            ],
            y=[0, 1, 0],
        )

    Parameters
    ----------

    x : Sequence
        Sequence of node features.
    edge_index : Sequence
        Edges represented via a graph connectivity format - matrix ``[[...src], [...dst]]``.
    y : Union[Sequence, float, int]
        Sequence of labels of all nodes or one graph label.
    edge_attr : Sequence | None
        Optional sequence of edge features. Default: ``None``
    y_mask : Sequence | None
        Optional sequence of node ids to generate queries for. Default: ``None`` (all nodes)

    """

    def __init__(
        self,
        x: Sequence,
        edge_index: Sequence,
        y: Union[Sequence, float, int] = 0.0,
        edge_attr: Sequence | None = None,
        y_mask: Sequence | None = None,
    ):
        self.x = x
        self.edge_index = edge_index
        self.y = y
        self.edge_attr = edge_attr
        self.y_mask = y_mask

    @staticmethod
    def get_query(
        y, output_name: str = "predict", one_hot_encode_labels: bool = False, max_classes=1, index: int | None = None
    ):
        relation = Relation.get(output_name)

        if index is not None:
            relation = relation(index)

        if one_hot_encode_labels:
            vector = y

            if not isinstance(y, (Sequence, np.ndarray)) or len(y) != max_classes:
                vector = np.zeros((max_classes,))
                vector[y[0] if isinstance(y, (Sequence, np.ndarray)) else y] = 1
            query = relation[vector]
        elif not isinstance(y, Iterable) or (getattr(y, "shape", -1) in (1, tuple())):
            query = relation[float(y)]
        elif len(y) == 1 and not isinstance(y[0], Sequence):
            query = relation[float(y[0])]
        else:
            if isinstance(y, (Sequence, np.ndarray)):
                query = relation[y]
            else:
                query = relation[y.detach().numpy()]
        return query

    @staticmethod
    def _get_edge_rel(edge_name, u, v, w):
        return Relation.get(edge_name)(int(u), int(v))[w].fixed()

    @staticmethod
    def _get_edge_rel_decode(edge_name, u, v, w):
        return Relation.get(f"{edge_name}_{np.argmax(w)}")(int(u), int(v))[1].fixed()

    def to_logic_form(
        self,
        feature_name: str = "node_feature",
        edge_name: str = "edge",
        output_name: str = "predict",
        one_hot_encode_labels=False,
        one_hot_decode_features=False,
        one_hot_decode_edge_features=False,
        max_classes=1,
    ) -> Tuple:
        if self.y_mask is not None:
            query = []

            for i in self.y_mask:
                query.append(Data.get_query(self.y[int(i)], output_name, one_hot_encode_labels, max_classes, int(i)))
        else:
            query = Data.get_query(self.y, output_name, one_hot_encode_labels, max_classes)

        edge_process = Data._get_edge_rel
        if one_hot_decode_edge_features:
            edge_process = Data._get_edge_rel_decode

        if self.edge_attr is None:
            example = [
                Relation.get(edge_name)(int(u), int(v))[1].fixed()
                for u, v in zip(self.edge_index[0], self.edge_index[1])
            ]
        elif isinstance(self.edge_attr, (Sequence, np.ndarray)):
            example = [
                edge_process(edge_name, u, v, w)
                for u, v, w in zip(self.edge_index[0], self.edge_index[1], self.edge_attr)
            ]
        else:
            example = [
                edge_process(edge_name, u, v, w)
                for u, v, w in zip(self.edge_index[0], self.edge_index[1], self.edge_attr.detach().numpy())
            ]

        if one_hot_decode_features:
            for i, features in enumerate(self.x):
                class_ = np.argmax(features)
                example.append(Relation.get(f"{feature_name}_{class_}")(i)[1].fixed())
        else:
            if isinstance(self.x, np.ndarray):
                for i, features in enumerate(self.x):
                    example.append(
                        Relation.get(feature_name)(i)[features[0] if features.size == 1 else features].fixed()
                    )
            elif isinstance(self.x, list):
                for i, features in enumerate(self.x):
                    example.append(
                        Relation.get(feature_name)(i)[
                            features[0] if len(features) == 1 and isinstance(features[0], (int, float)) else features
                        ].fixed()
                    )
            else:
                for i, features in enumerate(self.x):
                    weight = features.detach().numpy()
                    example.append(Relation.get(feature_name)(i)[weight[0] if weight.size == 1 else weight].fixed())
        return query, example

    @staticmethod
    def from_pyg(data) -> List["Data"]:
        """
        Converts a PyTorch Geometric Data instance into a list of PyNeuraLogic :py:class:`~Data` instances.
        The conversion supports :code:`train_mask`, :code:`test_mask` and :code:`val_mask` attributes -
        for each mask the conversion yields a new data instance.

        :param data: The PyTorch Geometric Data instance
        :return: The list of PyNeuraLogic :py:class:`~neuralogic.dataset.tensor.Data` instances
        """
        data_list = []

        if hasattr(data, "train_mask"):
            train_mask_index = data.train_mask.nonzero(as_tuple=False).view(-1)
            data_list.append(Data(data.x, data.edge_index, data.y, data.edge_attr, train_mask_index))
        if hasattr(data, "test_mask"):
            test_mask_index = data.test_mask.nonzero(as_tuple=False).view(-1)
            data_list.append(Data(data.x, data.edge_index, data.y, data.edge_attr, test_mask_index))
        if hasattr(data, "val_mask"):
            val_mask_index = data.val_mask.nonzero(as_tuple=False).view(-1)
            data_list.append(Data(data.x, data.edge_index, data.y, data.edge_attr, val_mask_index))
        if len(data_list) == 0:
            data_list.append(Data(data.x, data.edge_index, data.y, data.edge_attr, None))

        return data_list


class TensorDataset(ConvertibleDataset):
    r"""The ``TensorDataset`` holds a list of :py:class:`~neuralogic.dataset.tensor.Data` instances -
    a list of graphs represented in a tensor format.

    Parameters
    ----------

    data : List[Data]
        List of data (graph) instances.
    one_hot_encode_labels : bool
        Turn numerical labels into one hot encoded vectors - e.g., label ``2`` would be turned
        into a vector ``[0, 0, 1, .., 0]`` of length ``number_of_classes``.
        Default: ``False``
    one_hot_decode_features : bool = False
        Turn one hot encoded feature vectors into a scalar - e.g., feature vector ``[0, 0, 1]`` would be turned into
        a predicate ``<feature_name>_2``.
        Default: ``False``
    one_hot_decode_edge_features : bool = False
        Turn one hot encoded edge feature vectors into a scalar - e.g., edge feature vector ``[0, 0, 1]`` would be turned into
        a predicate ``<edge_name>_2``.
        Default: ``False``
    number_of_classes : int
        Specifies the number of classes for converting numerical labels to one hot encoded vectors.
        Default: ``1``
    feature_name : str
        Specify the node feature predicate name used for converting into the logic format.
        Default: ``"node_feature"``
    edge_name : str
        Specify the edge predicate name used for converting into the logic format.
        Default: ``"edge"``
    output_name : str
        Specify the output predicate name used for converting into the logic format.
        Default: ``"predict"``

    """

    def __init__(
        self,
        data: List[Data],
        one_hot_encode_labels: bool = False,
        one_hot_decode_features: bool = False,
        one_hot_decode_edge_features: bool = False,
        number_of_classes: int = 1,
        feature_name: str = "node_feature",
        edge_name: str = "edge",
        output_name: str = "predict",
    ):
        self.data = data

        self.one_hot_decode_features = one_hot_decode_features
        self.one_hot_decode_edge_features = one_hot_decode_edge_features
        self.one_hot_encode_labels = one_hot_encode_labels
        self.number_of_classes = number_of_classes

        self.feature_name = feature_name
        self.edge_name = edge_name
        self.output_name: str = output_name

    def add_data(self, data: Data):
        self.data.append(data)

    def to_dataset(self) -> Dataset:
        dataset = Dataset()

        for data in self.data:
            query, examples = data.to_logic_form(
                self.feature_name,
                self.edge_name,
                self.output_name,
                self.one_hot_encode_labels,
                self.one_hot_decode_features,
                self.one_hot_decode_edge_features,
                self.number_of_classes,
            )

            if isinstance(query, Sequence):
                for q in query:
                    dataset.add(q, examples)
            else:
                dataset.add(query, examples)
        return dataset

    def dump(
        self,
        queries_fp,
        examples_fp,
        sep: str = "\n",
    ):
        for data in self.data:
            query, examples = data.to_logic_form(
                self.feature_name,
                self.edge_name,
                self.output_name,
                self.one_hot_encode_labels,
                self.one_hot_decode_features,
                self.one_hot_decode_edge_features,
                self.number_of_classes,
            )

            queries_fp.write(f"{query}{sep}")
            examples_fp.write(f"{','.join(example.to_str(False) for example in examples)}.{sep}")



from typing import Union, Sequence

import jpype

from neuralogic.core.constructs.java_objects import JavaFactory
from neuralogic.core.constructs.relation import BaseRelation
from neuralogic.core.constructs.rule import Rule
from neuralogic.dataset.base import BaseDataset
from neuralogic.core.constructs.factories import R

DatasetEntries = Union[BaseRelation, Rule]


class Sample:
    __slots__ = (
        "query",
        "example",
    )

    def __init__(self, query: BaseRelation | list[BaseRelation] | None, example: Sequence[DatasetEntries] | DatasetEntries | None):
        self.query = query

        if example is None:
            example = []

        if not isinstance(example, Sequence):
            self.example = [example]
        else:
            self.example = example

    def draw(*args, **kwargs):
        raise NotImplementedError("sample cannot be drawn unless it is grounded or neuralized")

    def __str__(self) -> str:
        if isinstance(self.query, list):
            return ", ".join(str(q) for q in self.query)
        return str(self.query)

    def __len__(self) -> int:
        if self.example is None:
            return 0
        return len(self.example)


class Dataset(BaseDataset):
    r"""
    Dataset encapsulating (learning) samples in the form of logic format, allowing users to fully take advantage of the
    PyNeuraLogic library.
    """

    __slots__ = ("samples", "_examples", "_queries")

    def __init__(self, samples: list[Sample] | Sample | None = None):
        self.samples = []

        if isinstance(samples, list):
            self.samples = samples
        elif not isinstance(samples, list) and samples is not None:
            self.samples = [samples]

        self._examples: list[list[DatasetEntries]] = []
        self._queries: list[BaseRelation] = []

    def set_samples(self, samples: list[Sample]):
        self.samples = samples

    def add_samples(self, samples: list[Sample]) -> "Dataset":
        self.samples.extend(samples)

        return self

    def add_sample(self, sample: Sample) -> "Dataset":
        self.samples.append(sample)

        return self

    def add(self, query: BaseRelation | list[BaseRelation] | None, example: list[DatasetEntries] | None) -> "Dataset":
        self.samples.append(Sample(query, example))

        return self

    def __getitem__(self, item: int) -> Sample:
        return self.samples[item]

    def __setitem__(self, key: int, value: Sample):
        self.samples[key] = value

    def __delitem__(self, key: int):
        del self.samples[key]

    def __str__(self):
        return ". ".join(str(s) for s in self.samples)

    def __len__(self):
        return len(self.samples)

    # Deprecated
    def add_example(self, example):
        self.add_examples([example])

    def add_examples(self, examples: list):
        self._examples.extend(examples)

    def add_query(self, query):
        self.add_queries([query])

    def add_queries(self, queries: list):
        self._queries.extend(queries)

    def set_examples(self, examples: list):
        self._examples = examples

    def set_queries(self, queries: list):
        self._queries = queries

    def generate_features(self, feature_depth: int = 1, count_groundings: bool = True):
        java_factory = JavaFactory()

        clauses = []
        vertex_lit = R.get("__vert")
        vertex_lit.predicate.special = False
        vertex_lit.predicate.hidden = False

        for sample in self.samples:
            vertex = set()

            for e in sample.example:
                if isinstance(e, Rule):
                    vertex.update(self._get_constants(e.head))

                    for rel in e.body:
                        vertex.update(self._get_constants(rel))
                if isinstance(e, BaseRelation):
                    vertex.update(self._get_constants(e))

            example = [vertex_lit(vert) for vert in vertex]
            example.extend(sample.example)

            clauses.append(java_factory.to_clause(example))

        clause = jpype.java.util.ArrayList(clauses)

        namespace = "cz.cvut.fel.ida.logic.features.generation"

        jpype.JClass(f"{namespace}.FeatureGenerationSettings").COUNT_GROUNDINGS = count_groundings
        features = jpype.JClass(f"{namespace}.FeatureGenerator").generateFeatures(clause, feature_depth)

        table = [[int(i) for i in feats] for feats in features.table]
        clauses = [str(clause) for clause in features.features]

        return table, clauses

    @staticmethod
    def _get_constants(relation: BaseRelation):
        return [term for term in relation.terms if not str(relation)[0].isupper()]



import numpy as np
import pytest
import torch

from neuralogic.core import Template, Settings, V, R
from neuralogic.dataset import Dataset, Sample
from neuralogic.nn.module import MultiheadAttention


@pytest.mark.parametrize(
    "qdim, kdim, vdim, num_heads, sequence_len",
    [
        (10, 12, 14, 1, 3),  # One head
        (10, 12, 14, 2, 3),  # Two heads
    ],
)
def test_multiheadattention(qdim: int, kdim: int, vdim: int, num_heads: int, sequence_len: int):
    keys = torch.rand((sequence_len, kdim))
    queries = torch.rand((sequence_len, qdim))
    values = torch.rand((sequence_len, vdim))

    mha = torch.nn.MultiheadAttention(qdim, num_heads, bias=False, kdim=kdim, vdim=vdim)

    template = Template()
    template += MultiheadAttention(qdim, num_heads, "out", "q", "k", "v", vdim=vdim, kdim=kdim)

    model = template.build(Settings(iso_value_compression=False, chain_pruning=False))

    params = model.parameters()
    torch_params = list(mha.parameters())

    params["weights"][1] = torch_params[0].detach().numpy().tolist()
    params["weights"][2] = torch_params[2].detach().numpy().tolist()
    params["weights"][3] = torch_params[1].detach().numpy().tolist()
    params["weights"][4] = torch_params[3].detach().numpy().tolist()

    model.load_state_dict(params)

    example = []
    for i in range(sequence_len):
        example.append(R.v(i)[values[i]])
        example.append(R.k(i)[keys[i]])
        example.append(R.q(i)[queries[i]])

    dataset = Dataset([Sample(R.out(i), example) for i in range(sequence_len)])
    built_dataset = model.build_dataset(dataset)

    torch_results = mha(queries, keys, values)
    pyneuralogic_results = model.test(built_dataset)

    for t_res, pnl_res in zip(torch_results[0], pyneuralogic_results):
        assert np.allclose(t_res.detach().numpy(), np.array(pnl_res))



from neuralogic.core import Template, V, R


def test_neq():
    template = Template()
    template += R.head(V.X, V.Y) <= (R.special.neq(V.X, V.Y), R.val(V.X), R.val(V.Y))

    examples = [R.val(1), R.val(-1)]

    out = template.q(R.head(V.X, V.Y), examples)
    out = sorted(list(out), key=lambda a: a["X"] + a["Y"])

    assert len(out) == 2

    assert out[0]["X"] == "-1"
    assert out[0]["Y"] == "1"

    assert out[1]["X"] == "1"
    assert out[1]["Y"] == "-1"


def test_leq():
    template = Template()
    template += R.head(V.X, V.Y) <= (R.special.leq(V.X, V.Y), R.val(V.X), R.val(V.Y))

    examples = [R.val(1), R.val(-1)]

    out = template.q(R.head(V.X, V.Y), examples)
    out = sorted(list(out), key=lambda a: a["X"] + a["Y"])

    assert len(out) == 3

    assert out[0]["X"] == "-1"
    assert out[0]["Y"] == "-1"

    assert out[1]["X"] == "-1"
    assert out[1]["Y"] == "1"

    assert out[2]["X"] == "1"
    assert out[2]["Y"] == "1"


def test_geq():
    template = Template()
    template += R.head(V.X, V.Y) <= (R.special.geq(V.X, V.Y), R.val(V.X), R.val(V.Y))

    examples = [R.val(1), R.val(-1)]

    out = template.q(R.head(V.X, V.Y), examples)
    out = sorted(list(out), key=lambda a: a["X"] + a["Y"])

    assert len(out) == 3

    assert out[0]["X"] == "-1"
    assert out[0]["Y"] == "-1"

    assert out[1]["X"] == "1"
    assert out[1]["Y"] == "-1"

    assert out[2]["X"] == "1"
    assert out[2]["Y"] == "1"


def test_lt():
    template = Template()
    template += R.head(V.X, V.Y) <= (R.special.lt(V.X, V.Y), R.val(V.X), R.val(V.Y))

    examples = [R.val(1), R.val(-1)]

    out = template.q(R.head(V.X, V.Y), examples)
    out = sorted(list(out), key=lambda a: a["X"] + a["Y"])

    assert len(out) == 1

    assert out[0]["X"] == "-1"
    assert out[0]["Y"] == "1"


def test_gt():
    template = Template()
    template += R.head(V.X, V.Y) <= (R.special.gt(V.X, V.Y), R.val(V.X), R.val(V.Y))

    examples = [R.val(1), R.val(-1)]

    out = template.q(R.head(V.X, V.Y), examples)
    out = sorted(list(out), key=lambda a: a["X"] + a["Y"])

    assert len(out) == 1

    assert out[0]["X"] == "1"
    assert out[0]["Y"] == "-1"


def test_eq():
    template = Template()
    template += R.head(V.X, V.Y) <= (R.special.eq(V.X, V.Y), R.val(V.X), R.val(V.Y))

    examples = [R.val(1), R.val(-1)]

    out = template.q(R.head(V.X, V.Y), examples)
    out = sorted(list(out), key=lambda a: a["X"] + a["Y"])

    assert len(out) == 2

    assert out[0]["X"] == "-1"
    assert out[0]["Y"] == "-1"

    assert out[1]["X"] == "1"
    assert out[1]["Y"] == "1"


def test_next():
    template = Template()
    template += R.head(V.X, V.Y) <= (R.special.next(V.X, V.Y), R.val(V.X), R.val(V.Y))

    examples = [R.val(1), R.val(-1), R.val(0)]

    out = template.q(R.head(V.X, V.Y), examples)
    out = sorted(list(out), key=lambda a: a["X"] + a["Y"])

    assert len(out) == 2

    assert out[0]["X"] == "-1"
    assert out[0]["Y"] == "0"

    assert out[1]["X"] == "0"
    assert out[1]["Y"] == "1"


def test_next_skip():
    template = Template()
    template += R.head(V.X, V.Y) <= (R.special.next(V.X, V.Z), R.special.next(V.Z, V.Y), R.val(V.X), R.val(V.Y))

    examples = [R.val(1), R.val(-1), R.val(0)]

    out = template.q(R.head(V.X, V.Y), examples)
    out = sorted(list(out), key=lambda a: a["X"] + a["Y"])

    assert len(out) == 1

    assert out[0]["X"] == "-1"
    assert out[0]["Y"] == "1"


def test_alldiff():
    template = Template()
    template += R.head(V.X, V.Y, V.Z) <= (R.special.alldiff(V.X, V.Y, V.Z), R.val(V.X), R.val(V.Y), R.val(V.Z))

    examples = [R.val(1), R.val(-1), R.val(0)]

    out = template.q(R.head(V.X, V.Y, V.Z), examples)
    out = sorted(list(out), key=lambda a: a["X"] + a["Y"])

    assert len(out) == 6



from typing import List

from neuralogic import manual_seed
from neuralogic.core import Settings, Template
from neuralogic.dataset.base import BaseDataset
from neuralogic.optim import SGD
from neuralogic.utils.data import XOR, XOR_Vectorized, Trains, Mutagenesis

from examples.datasets import (
    multiple_examples_trains,
    multiple_examples_no_order_trains,
    naive_trains,
    naive_xor,
    horses,
    vectorized_xor,
)


import pytest


@pytest.mark.parametrize(
    "template, dataset, expected_results",
    [
        (*XOR(), [0, 0.625, 0.645, 0.663]),
        (*XOR_Vectorized(), [0, 0.7, 0.657, -0.056]),
        (
            *Trains(),
            [0.755, 0.75, -0.761, 0.749, 0.76, 0.724, -0.753, 0.754, -0.748, -0.755, -0.757, -0.761, 0.73, 0.742,
             -0.752, -0.691, 0.748, 0.758, -0.73, -0.75]
        ),
        (
            *Mutagenesis(),
            [
                0.473,
                0.358,
                -0.083,
                0.069,
                -0.022,
                0.221,
                0.221,
                -0.014,
                0.216,
                -0.052,
                0.11,
                0.065,
                0.029,
                0.029,
                0.116,
                -0.008,
                0.255,
                -0.214,
                0.086,
                0.295,
                0.473,
                0.07,
                0.114,
                0.221,
                0.255,
                0.141,
                0.266,
                0.019,
                -0.106,
                -0.014,
                0.022,
                0.388,
                0.202,
                0.221,
                -0.14,
                0.316,
                -0.047,
                0.425,
                0.377,
                0.376,
                0.292,
                0.316,
                0.377,
                0.473,
                0.473,
                0.255,
                0.047,
                0.227,
                0.429,
                0.469,
                -0.045,
                -0.052,
                0.469,
                0.266,
                -0.015,
                0.021,
                -0.166,
                -0.084,
                -0.09,
                0.086,
                0.329,
                0.064,
                0.358,
                0.316,
                0.007,
                0.221,
                0.468,
                0.256,
                0.316,
                0.429,
                0.406,
                0.285,
                0.137,
                0.096,
                0.607,
                0.303,
                0.473,
                0.256,
                0.171,
                0.011,
                0.022,
                0.112,
                0.033,
                0.135,
                0.473,
                -0.147,
                0.26,
                -0.014,
                -0.09,
                0.141,
                0.381,
                0.303,
                0.266,
                0.086,
                0.267,
                0.429,
                0.134,
                -0.083,
                0.261,
                0.358,
                -0.015,
                0.115,
                0.316,
                0.129,
                0.015,
                0.221,
                0.26,
                0.379,
                -0.14,
                0.055,
                -0.083,
                0.309,
                0.468,
                0.266,
                0.377,
                0.406,
                0.267,
                0.224,
                0.477,
                -0.046,
                0.266,
                0.216,
                -0.008,
                0.075,
                -0.047,
                -0.103,
                0.128,
                -0.105,
                -0.038,
                -0.026,
                -0.106,
                -0.143,
                0.115,
                0.029,
                -0.12,
                -0.143,
                -0.194,
                -0.072,
                -0.4,
                0.029,
                0.038,
                -0.098,
                0.029,
                -0.149,
                -0.162,
                -0.136,
                -0.05,
                -0.118,
                -0.036,
                -0.093,
                0.084,
                -0.068,
                -0.013,
                0.042,
                0.029,
                0.058,
                0.061,
                -0.013,
                -0.052,
                0.029,
                -0.17,
                -0.048,
                -0.045,
                -0.087,
                0.06,
                -0.162,
                -0.196,
                -0.068,
                0.174,
                -0.048,
                0.02,
                -0.013,
                -0.161,
                0.01,
                -0.137,
                -0.107,
                -0.068,
                -0.143,
                0.072,
                -0.162,
                -0.065,
                0.174,
                -0.058,
                0.133,
                -0.086,
                -0.031,
                0.065,
                -0.025,
            ],
        ),
    ],
)
def test_evaluator_run_on_files(template: Template, dataset: BaseDataset, expected_results: List[float]) -> None:
    """Tests for running java evaluator on files"""
    manual_seed(0)
    settings = Settings(optimizer=SGD(0.1))

    template.build(settings)

    built_dataset = template.build_dataset(dataset)
    template.train(built_dataset, epochs=50)

    results = []
    for predicted in template.test(built_dataset):
        results.append(round(predicted, 3))

    assert len(results) == len(expected_results)

    for result, expected_result in zip(results, expected_results):
        assert result == expected_result


@pytest.mark.parametrize(
    "template, dataset, expected_results, seed",
    [
        (naive_xor.template, naive_xor.dataset, [0, 0.936, 0.935, -0.002], 0),
        (vectorized_xor.template, vectorized_xor.dataset, [0, 0.955, 0.954, -0.003], 0),
        (horses.template, horses.dataset, [0.951, 0], 0),
        (
            naive_trains.template,
            naive_trains.dataset,
            [0.76, 0.761, -0.762, 0.761, 0.761, 0.758, -0.76, 0.761, -0.76, -0.761, -0.761, -0.761, 0.757, 0.757,
             -0.761, -0.754, 0.759, 0.761, -0.757, -0.761],
            0,
        ),
        (
            multiple_examples_trains.template,
            multiple_examples_trains.dataset,
            [
                0.743,
                0.745,
                0.735,
                -0.761,
                -0.74,
                -0.761,
                0.739,
                0.761,
                -0.761,
                -0.761,
                -0.761,
                -0.761,
                -0.761,
                -0.761,
                -0.746,
                -0.761,
                -0.761,
                -0.749,
                -0.734,
                0.761,
            ],
            0,
        ),
        (
            multiple_examples_no_order_trains.template,
            multiple_examples_no_order_trains.dataset,
            [
                0.685,
                0.715,
                0.746,
                0.759,
                0.556,
                0.731,
                0.128,
                0.724,
                0.158,
                0.751,
                -0.699,
                -0.761,
                -0.759,
                0.761,
                -0.762,
                -0.756,
                -0.755,
                -0.761,
                -0.762,
                -0.754,
            ],
            1,
        ),
    ],
)
def test_evaluator_run_on_rules(
    template: Template, dataset: BaseDataset, expected_results: List[float], seed: int
) -> None:
    """Tests for running java evaluator on rules"""
    manual_seed(seed)
    settings = Settings(optimizer=SGD(lr=0.1))

    template.build(settings)

    built_dataset = template.build_dataset(dataset)
    template.train(built_dataset, epochs=300)

    results = []
    for predicted in template.test(built_dataset):
        results.append(round(predicted, 3))

    assert len(results) == len(expected_results)

    for result, expected_result in zip(results, expected_results):
        assert result == expected_result


@pytest.mark.parametrize(
    "template, dataset",
    [
        (naive_xor.template, naive_xor.dataset),
    ],
)
def test_evaluator_state_loading(template: Template, dataset: BaseDataset) -> None:
    """Tests for loading state"""
    settings = Settings(optimizer=SGD(0.1), epochs=20)

    template.build(settings)
    built_dataset = template.build_dataset(dataset)
    template.train(built_dataset)

    results = []
    for predicted in template.test(built_dataset):
        results.append(round(predicted, 5))

    second_template = template.clone()
    second_template.build(settings)

    built_dataset = second_template.build_dataset(dataset)

    second_results = []
    for predicted in second_template.test(built_dataset):
        second_results.append(round(predicted, 5))

    assert len(results) == len(second_results)
    assert any(result != second_result for result, second_result in zip(results, second_results))

    second_template.load_state_dict(template.state_dict())

    second_results = []
    for predicted in second_template.test(built_dataset):
        second_results.append(round(predicted, 5))

    assert len(results) == len(second_results)
    for result, second_result in zip(results, second_results):
        assert result == second_result



from neuralogic.core.constructs.factories import Var, Const, Relation, V, C, R
from neuralogic.core.constructs.rule import Rule, RuleBody
from neuralogic.core.template import Template
from neuralogic.core.builder import BuiltDataset, GroundedDataset
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.settings import Settings, SettingsProxy
from neuralogic.core.enums import Grounder
from neuralogic.core.constructs.function import Transformation, Aggregation, Combination, F


__all__ = [
    "Var",
    "V",
    "Const",
    "C",
    "Relation",
    "R",
    "F",
    "Rule",
    "RuleBody",
    "Template",
    "BuiltDataset",
    "GroundedDataset",
    "Transformation",
    "Aggregation",
    "Combination",
    "Grounder",
    "Settings",
    "SettingsProxy",
    "Metadata",
]



class BaseDataset:
    def dump(
        self,
        queries_fp,
        examples_fp,
        sep: str = "\n",
    ):
        raise NotImplementedError


class ConvertibleDataset(BaseDataset):
    def to_dataset(self):
        raise NotImplementedError



from neuralogic.core.constructs.function.function import TransformationFunction
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation, Aggregation, Combination
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class GATv2Conv(Module):
    r"""
    GATv2 layer from `"How Attentive are Graph Attention Networks?" <https://arxiv.org/abs/2105.14491>`_.

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : str
        Edge predicate name to use for neighborhood relations.
    share_weights : bool
        Share weights in attention. Default: ``False``
    activation : TransformationFunction
        Activation function of the output.
        Default: ``Transformation.IDENTITY``

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: str,
        share_weights: bool = False,
        activation: TransformationFunction = Transformation.IDENTITY,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.share_weights = share_weights
        self.activation = activation

    def __call__(self):
        w1 = f"{self.output_name}__right"
        w2 = w1 if self.share_weights else f"{self.output_name}__left"

        attention = R.get(f"{self.output_name}__attention")
        attention_metadata = Metadata(transformation=Transformation.LEAKY_RELU)
        metadata = Metadata(aggregation=Aggregation.SUM, combination=Combination.PRODUCT)

        head = R.get(self.output_name)
        feature = R.get(self.feature_name)
        edge = R.get(self.edge_name)

        return [
            (
                attention(V.I, V.J)[self.out_channels, self.out_channels]
                <= (
                    feature(V.I)[w2 : self.out_channels, self.in_channels],
                    feature(V.J)[w1 : self.out_channels, self.in_channels],
                )
            )
            | attention_metadata,
            attention / 2 | Metadata(transformation=Transformation.SOFTMAX),
            (head(V.I) <= (attention(V.I, V.J), feature(V.J)[w1 : self.out_channels, self.in_channels], edge(V.J, V.I)))
            | metadata,
            head / 1 | Metadata(transformation=self.activation),
        ]



from typing import Union, Tuple, Optional

import jpype

from neuralogic.core.constructs.function.function import TransformationFunction


class Reshape(TransformationFunction):
    __slots__ = ("shape",)

    def __init__(
        self,
        name: str,
        *,
        shape: Tuple[int, int] | int | None = None,
    ):
        super().__init__(name)

        self.shape = (shape,) if isinstance(shape, int) else shape

    def __call__(
        self,
        relation: Optional = None,
        *,
        shape: Union[None, Tuple[int, int], int] = None,
    ):
        reshape = Reshape(self.name, shape=shape)
        return TransformationFunction.__call__(reshape, relation)

    def is_parametrized(self) -> bool:
        return True

    def get(self):
        shape = None if self.shape is None else list(self.shape)

        return jpype.JClass("cz.cvut.fel.ida.algebra.functions.transformation.joint.Reshape")(shape)

    def wrap(self, content: str) -> str:
        return f"reshape({content}, shape={self.shape})"

    def __str__(self):
        return f"reshape(shape={self.shape})"



from typing import List, Union, Sequence

from neuralogic.core.constructs.function.function import TransformationFunction
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation
from neuralogic.core.constructs.factories import R
from neuralogic.nn.module.module import Module


class MLP(Module):
    r"""

    Parameters
    ----------

    units : List[int]
        List of layer sizes.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input name.
    activation : Union[TransformationFunction, List[TransformationFunction]]
        Activation function of all layers or list of activations for each layer.
        Default: ``Transformation.RELU``
    arity : int
        Default: ``-1``
    """

    def __init__(
        self,
        units: List[int],
        output_name: str,
        input_name: str,
        activation: Union[TransformationFunction, List[TransformationFunction]] = Transformation.RELU,
        arity: int = 1,
    ):
        self.output_name = output_name
        self.input_name = input_name

        self.units = units
        self.activation: Union[TransformationFunction, List[TransformationFunction]] = activation
        self.arity = arity

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity)]
        layers = []

        prev_layer = R.get(self.input_name)
        iters = len(self.units) - 1

        if isinstance(self.activation, Sequence):
            metadata = [Metadata(transformation=act) for act in self.activation]
            metadata.extend([Metadata()] * (iters - len(metadata)))
        else:
            metadata = [Metadata(transformation=self.activation)] * (iters + 1)

        for index in range(0, iters, 2):
            in_channels, out_channels = self.units[index], self.units[index + 1]

            head_predicate = R.get(f"{self.output_name}__mlp{index // 2}")
            body_literal = prev_layer(terms)[out_channels, in_channels]
            head_literal = head_predicate(terms)

            body_metadata = metadata[index]
            head_metadata = metadata[index + 1]

            if body_metadata is None:
                if index < len(self.activation):
                    body_metadata = Metadata(transformation=self.activation[index])
                else:
                    body_metadata = []

            if index + 2 < len(self.units):
                in_channels, out_channels = self.units[index + 1], self.units[index + 2]
                head_literal = head_literal[out_channels, in_channels]

            layers.append((head_literal <= body_literal) | body_metadata)
            layers.append(head_predicate / self.arity | head_metadata)
            prev_layer = head_predicate

        layers[-2].head.predicate.name = self.output_name
        layers[-1].predicate.name = self.output_name

        return layers
